"""FHIRProvider class, for registering FHIR interactions with a FHIRStarter app."""

from typing import (
    Any,
    Callable,
    Iterable,
    List,
    Mapping,
    Protocol,
    Sequence,
    Type,
    TypeVar,
    Union,
)

from fastapi import params

from .fhir_specification import FHIR_SEQUENCE
from .interactions import (
    CreateInteraction,
    CreateInteractionHandler,
    DeleteInteraction,
    DeleteInteractionHandler,
    InteractionHandler,
    PatchInteraction,
    PatchInteractionHandler,
    ReadInteraction,
    ReadInteractionHandler,
    ResourceType,
    SearchTypeInteraction,
    SearchTypeInteractionHandler,
    TypeInteraction,
    UpdateInteraction,
    UpdateInteractionHandler,
)
from .resources import Resource

C = TypeVar("C", bound=Callable[..., Any])


class TypeInteractionType(Protocol[ResourceType]):
    @staticmethod
    def __call__(
        resource_type: Type[ResourceType],
        handler: InteractionHandler[ResourceType],
        route_options: Mapping[str, Any],
    ) -> TypeInteraction[ResourceType]: ...


class FHIRProvider:
    """
    Class that contains a collection of FHIR interactions to be added to a FHIRStarter app. These
    interactions are added as API routes.

    Aside from instantiation, interaction with this class is performed solely through the register
    decorators, e.g. register_read_interaction. One must use these decorators to decorate
    functions that perform FHIR interactions.
    """

    def __init__(
        self, *, dependencies: Union[Sequence[params.Depends], None] = None
    ) -> None:
        self._dependencies = dependencies or []
        self._interactions: List[TypeInteraction[Resource]] = []

    @property
    def interactions(self) -> Iterable[TypeInteraction[Resource]]:
        yield from self._interactions

    def read(
        self,
        resource_type: Type[ResourceType],
        *,
        dependencies: Union[Sequence[params.Depends], None] = None,
        include_in_schema: bool = True,
    ) -> Callable[
        [ReadInteractionHandler[ResourceType]],
        ReadInteractionHandler[ResourceType],
    ]:
        """Register a FHIR read interaction."""
        return self._register_type_interaction(
            resource_type,
            ReadInteraction[ResourceType],
            dependencies,
            include_in_schema,
        )

    def update(
        self,
        resource_type: Type[ResourceType],
        *,
        dependencies: Union[Sequence[params.Depends], None] = None,
        include_in_schema: bool = True,
    ) -> Callable[
        [UpdateInteractionHandler[ResourceType]], UpdateInteractionHandler[ResourceType]
    ]:
        """Register a FHIR update interaction."""
        return self._register_type_interaction(
            resource_type,
            UpdateInteraction[ResourceType],
            dependencies,
            include_in_schema,
        )

    def patch(
        self,
        resource_type: Type[ResourceType],
        *,
        dependencies: Union[Sequence[params.Depends], None] = None,
        include_in_schema: bool = True,
    ) -> Callable[
        [PatchInteractionHandler[ResourceType]], PatchInteractionHandler[ResourceType]
    ]:
        """Register a FHIR patch interaction."""
        return self._register_type_interaction(
            resource_type,
            PatchInteraction[ResourceType],
            dependencies,
            include_in_schema,
        )

    def delete(
        self,
        resource_type: Type[ResourceType],
        *,
        dependencies: Union[Sequence[params.Depends], None] = None,
        include_in_schema: bool = True,
    ) -> Callable[[DeleteInteractionHandler], DeleteInteractionHandler]:
        """Register a FHIR delete interaction."""
        return self._register_type_interaction(
            resource_type,
            DeleteInteraction[ResourceType],
            dependencies,
            include_in_schema,
        )

    def create(
        self,
        resource_type: Type[ResourceType],
        *,
        dependencies: Union[Sequence[params.Depends], None] = None,
        include_in_schema: bool = True,
    ) -> Callable[
        [CreateInteractionHandler[ResourceType]], CreateInteractionHandler[ResourceType]
    ]:
        """Register a FHIR create interaction."""
        return self._register_type_interaction(
            resource_type,
            CreateInteraction[ResourceType],
            dependencies,
            include_in_schema,
        )

    def search_type(
        self,
        resource_type: Type[ResourceType],
        *,
        dependencies: Union[Sequence[params.Depends], None] = None,
        include_in_schema: bool = True,
    ) -> Callable[[SearchTypeInteractionHandler], SearchTypeInteractionHandler]:
        """Register a FHIR search-type interaction."""
        return self._register_type_interaction(
            resource_type,
            SearchTypeInteraction[ResourceType],
            dependencies,
            include_in_schema,
        )

    def _register_type_interaction(
        self,
        resource_type: Type[ResourceType],
        type_interaction_cls: TypeInteractionType[ResourceType],
        dependencies: Union[Sequence[params.Depends], None],
        include_in_schema: bool,
    ) -> Callable[[C], C]:
        _check_resource_type_module(resource_type)

        def decorator(handler: C) -> C:
            self._interactions.append(
                type_interaction_cls(
                    resource_type=resource_type,
                    handler=handler,
                    route_options={
                        "dependencies": (*self._dependencies, *(dependencies or ())),
                        "include_in_schema": include_in_schema,
                    },
                )
            )
            return handler

        return decorator


def _check_resource_type_module(resource_type: Type[Resource]) -> None:
    """Ensure that the resource type is compatible with the server's defined FHIR sequence."""

    # Get the module name of the resource's fhir.resources parent class. If a user is using a model
    # with custom examples, then their model will inherit from something that this code will
    # recognize.
    #
    # Note: This limitation currently disallows a deeper inheritance hierarchy.
    module = ""
    bases = resource_type.__bases__
    for base in bases:
        if base.__module__.startswith("fhir.resources"):
            module = base.__module__
    assert (
        module
    ), f"Unable to determine FHIR sequence of resource {resource_type.get_resource_type()}"

    if FHIR_SEQUENCE in ("R4", "R4"):
        assert not module.startswith(
            ("fhir.resources.STU3", "fhir.resources.R4B")
        ), f"Resource types from {module} cannot be used with FHIR sequence {FHIR_SEQUENCE}"
    elif FHIR_SEQUENCE == "STU3":
        assert module.startswith(
            "fhir.resources.STU3"
        ), "Only resource types from fhir.resources.STU3 can be used with FHIR sequence STU3"
    elif FHIR_SEQUENCE == "R4B":
        assert module.startswith(
            "fhir.resources.R4B"
        ), "Only resource types from fhir.resources.R4B can be used with FHIR sequence R4B"



"""
Resources module that finesses imports of resources used by the framework so that they can be
referenced by the same names across the framework regardless of the FHIR version being used.

The finesse required here is due to what is supported and not supported by various versions of the
fhir.resources package. Versions including and after 7.0.0 no longer support the R4 sequence, but do
support the STU3, R4B, and R5 sequences. Version 6.5.0 supports the R4 sequence.

Due to this, the allowed versions of the fhir.resources package will be restricted based on the FHIR
sequence being used.
"""

from .fhir_specification import FHIR_SEQUENCE

if FHIR_SEQUENCE in ("R4", "R5"):
    from fhir.resources.bundle import Bundle
    from fhir.resources.capabilitystatement import CapabilityStatement
    from fhir.resources.fhirtypes import Id
    from fhir.resources.operationoutcome import OperationOutcome
    from fhir.resources.resource import Resource
elif FHIR_SEQUENCE == "STU3":
    from fhir.resources.STU3.bundle import Bundle
    from fhir.resources.STU3.capabilitystatement import CapabilityStatement
    from fhir.resources.STU3.fhirtypes import Id
    from fhir.resources.STU3.operationoutcome import OperationOutcome
    from fhir.resources.STU3.resource import Resource
elif FHIR_SEQUENCE == "R4B":
    from fhir.resources.R4B.bundle import Bundle
    from fhir.resources.R4B.capabilitystatement import CapabilityStatement
    from fhir.resources.R4B.fhirtypes import Id
    from fhir.resources.R4B.operationoutcome import OperationOutcome
    from fhir.resources.R4B.resource import Resource

__all__ = ["Bundle", "CapabilityStatement", "Id", "OperationOutcome", "Resource"]



"""
Standard exception types for reporting errors.

The exception classes defined here provide a method which will return an OperationOutcome.
"""

from abc import ABC, abstractmethod
from typing import Union

from fastapi import Request, status
from fastapi.exceptions import HTTPException

from .resources import OperationOutcome
from .utils import make_operation_outcome, parse_fhir_request


class FHIRException(HTTPException, ABC):
    """
    Abstract base class for all FHIR exceptions.

    This class provides a set_request method that provides concrete subclasses with the request
    object for additional context.
    """

    def __init__(self, status_code: int, details_text: Union[str, None] = None) -> None:
        super().__init__(status_code, detail=details_text)
        self._request: Union[Request, None] = None

    def set_request(self, request: Request) -> None:
        self._request = request

    @abstractmethod
    def operation_outcome(self) -> OperationOutcome:
        raise NotImplementedError


class FHIRHTTPException(FHIRException):
    """
    Abstract base class for FHIR errors that map neatly to OperationOutcome issue type codes.
    Example: HTTP status code 409 <-> OperationOutcome issue type code "conflict"

    If these mappings are not desired, FHIRGeneralError can be used to create the exact desired
    response.
    """

    _STATUS_CODE_MAPPINGS = {
        401: "unknown",
        403: "forbidden",
        405: "not-supported",
        406: "not-supported",
        409: "conflict",
        410: "deleted",
        412: "conflict",
        415: "not-supported",
    }

    def __init__(self, details_text: Union[str, None] = None):
        super().__init__(self._status_code(), details_text)

    def operation_outcome(self) -> OperationOutcome:
        return make_operation_outcome(
            severity="error",
            code=self._STATUS_CODE_MAPPINGS[self._status_code()],
            details_text=self.detail,
        )

    @classmethod
    @abstractmethod
    def _status_code(cls) -> int:
        raise NotImplementedError


class FHIRGeneralError(FHIRException):
    """
    General FHIR exception class.

    This class allows the caller to pass in a predefined OperationOutcome.
    """

    def __init__(
        self, status_code: int, severity: str, code: str, details_text: str
    ) -> None:
        super().__init__(status_code, details_text)
        self._operation_outcome_ = make_operation_outcome(severity, code, details_text)

    @classmethod
    def from_operation_outcome(
        cls, status_code: int, operation_outcome: OperationOutcome
    ) -> "FHIRGeneralError":
        error = FHIRGeneralError(status_code, "severity", "code", "details")
        error._operation_outcome_ = operation_outcome
        return error

    def operation_outcome(self) -> OperationOutcome:
        return self._operation_outcome_


class FHIRBadRequestError(FHIRException):
    """FHIR exception class for 400 bad request errors."""

    def __init__(self, code: str, details_text: str) -> None:
        super().__init__(status.HTTP_400_BAD_REQUEST, details_text)
        self._code = code

    def operation_outcome(self) -> OperationOutcome:
        return make_operation_outcome(
            severity="error", code=self._code, details_text=self.detail
        )


class FHIRUnauthorizedError(FHIRHTTPException):
    """FHIR exception class for 401 unauthorized errors."""

    @classmethod
    def _status_code(cls) -> int:
        return status.HTTP_401_UNAUTHORIZED


class FHIRForbiddenError(FHIRHTTPException):
    """FHIR exception class for 403 forbidden errors."""

    @classmethod
    def _status_code(cls) -> int:
        return status.HTTP_403_FORBIDDEN


class FHIRResourceNotFoundError(FHIRException):
    """FHIR exception class for 404 not found errors."""

    def __init__(self) -> None:
        super().__init__(status.HTTP_404_NOT_FOUND)

    def operation_outcome(self) -> OperationOutcome:
        try:
            interaction_info = parse_fhir_request(self._request)
        except Exception as exception:
            raise AssertionError(
                "Unable to get resource type and resource ID from request; request must be set "
                "before the response is created"
            ) from exception
        else:
            return make_operation_outcome(
                "error",
                "not-found",
                f"Unknown {interaction_info.resource_type} resource "
                f"'{interaction_info.resource_id}'",
            )


class FHIRMethodNotAllowedError(FHIRHTTPException):
    """FHIR exception class for 405 method not allowed errors."""

    @classmethod
    def _status_code(cls) -> int:
        return status.HTTP_405_METHOD_NOT_ALLOWED


class FHIRNotAcceptableError(FHIRHTTPException):
    """FHIR exception class for 406 not acceptable errors."""

    @classmethod
    def _status_code(cls) -> int:
        return status.HTTP_406_NOT_ACCEPTABLE


class FHIRConflictError(FHIRHTTPException):
    """FHIR exception class for 409 conflict errors."""

    @classmethod
    def _status_code(cls) -> int:
        return status.HTTP_409_CONFLICT


class FHIRGoneError(FHIRHTTPException):
    """FHIR exception class for 410 gone errors."""

    @classmethod
    def _status_code(cls) -> int:
        return status.HTTP_410_GONE


class FHIRPreconditionFailedError(FHIRHTTPException):
    """FHIR exception class for 412 precondition failed errors."""

    @classmethod
    def _status_code(cls) -> int:
        return status.HTTP_412_PRECONDITION_FAILED


class FHIRUnsupportedMediaTypeError(FHIRHTTPException):
    """FHIR exception class for 415 unsupported media type errors."""

    @classmethod
    def _status_code(cls) -> int:
        return status.HTTP_415_UNSUPPORTED_MEDIA_TYPE


class FHIRUnprocessableEntityError(FHIRException):
    """FHIR exception class for 422 unprocessable entity errors."""

    def __init__(self, code: str, details_text: str) -> None:
        super().__init__(status.HTTP_422_UNPROCESSABLE_ENTITY, details_text)
        self._code = code

    def operation_outcome(self) -> OperationOutcome:
        return make_operation_outcome(
            severity="error", code=self._code, details_text=self.detail
        )



from fastapi import Depends, Request, Response, status

from .fhir_specification import FHIR_SEQUENCE, FHIR_VERSION
from .fhirstarter import FHIRStarter
from .interactions import InteractionContext
from .json_patch import JSONPatch, convert_json_patch
from .providers import FHIRProvider
from .utils import is_resource_type, parse_fhir_request

__all__ = [
    "Depends",
    "FHIRProvider",
    "FHIRStarter",
    "FHIR_SEQUENCE",
    "FHIR_VERSION",
    "InteractionContext",
    "JSONPatch",
    "Request",
    "Response",
    "convert_json_patch",
    "is_resource_type",
    "parse_fhir_request",
    "status",
]



"""
Functions for working with search parameters defined for search operations in the FHIR
specification.
"""

import inspect
import re
from copy import deepcopy
from dataclasses import dataclass

try:
    from functools import cache
except ImportError:
    from functools import lru_cache as cache

from typing import Any, Callable, Dict, Mapping, Tuple, Union

from fastapi import Request, Response

from .fhir_specification.utils import (
    load_extra_search_parameters,
    load_search_parameters,
)
from .interactions import InteractionContext


class SearchParameters:
    def __init__(
        self,
        custom_search_parameters: Union[
            Mapping[str, Mapping[str, Mapping[str, str]]], None
        ] = None,
    ):
        self._custom_search_parameters = custom_search_parameters or {}

    def get_metadata(self, resource_type: str) -> Dict[str, Dict[str, str]]:
        """
        Return search parameter metadata for the given resource type.

        For a given resource type, the search parameter metadata is a union between the search
        parameter metadata for the resource type itself, DomainResource, Resource, and custom search
        parameter metadata.
        """
        all_search_parameters = _load_search_parameters_file()

        search_parameters = deepcopy(all_search_parameters.get(resource_type, {}))
        search_parameters.update(all_search_parameters["DomainResource"])
        search_parameters.update(all_search_parameters["Resource"])
        search_parameters.update(self._custom_search_parameters.get(resource_type, {}))

        return search_parameters


def var_name_to_qp_name(name: str) -> str:
    """
    Convert a Python-friendly variable name to a FHIR query parameter name.

    Search parameters for specific resources in FHIR are lowercase strings that may have dashes in
    them. The Python-friendly versions of them have underscores instead of dashes. Underscores may
    be at the end of the variable name as well, if they conflict with a Python reserved word.

    Search parameters for all resources in FHIR start with underscore, and are camelcase. The
    Python-friendly versions of these have their uppercase characters replaced with an underscore
    plus the lowercase version of the character.
    """
    if name.startswith("_"):
        return f"_{re.sub('_[a-z]', lambda m: m.group(0)[1:].upper(), name[1:])}"

    if name.endswith("_"):
        name = name[:-1]
    return name.replace("_", "-")


@dataclass
class SupportedSearchParameter:
    name: str
    multiple: bool


def supported_search_parameters(
    search_function: Callable[..., Any]
) -> Tuple[SupportedSearchParameter, ...]:
    """
    Given a callable, return a list of the parameter names in the function (excluding variadic
    keyword and variadic positional arguments).

    This function is used to determine what search parameters are supported by the handler supplied
    for a registered FHIR search interaction.
    """
    # TODO: There is probably a more sophisticated way of figuring out if list[str] is part of an
    #  annotation, but this is sufficient for now.
    return tuple(
        SupportedSearchParameter(
            name=name, multiple="list[str]" in str(parameter.annotation).lower()  # type: ignore[call-arg]
        )
        for name, parameter in inspect.signature(search_function).parameters.items()
        if parameter.annotation != InteractionContext
    )


def search_parameter_sort_key(
    name: str,
    search_parameter_metadata: Dict[str, Dict[str, str]],
    parameter_annotation: Union[type, None] = None,
) -> Tuple[bool, bool, bool, bool, bool, str]:
    """
    Return a sort key for a search parameter.

    This function allows for consistent sorting of search parameters throughout the package.

    Sort order is:
    1. Request and response annotations
    2. Parameters that do not start with underscore
    3. Parameters that are search or search result parameters (i.e. _format and _pretty go to the
       bottom)
    4. Parameters that are part of the capability statement (this favors search parameters like
       _has over search result parameters like _sort)
    5. Alphabetical by name

    """
    return (
        parameter_annotation != Request,
        parameter_annotation != Response,
        name.startswith("_"),
        var_name_to_qp_name(name) not in search_parameter_metadata,
        not search_parameter_metadata.get(var_name_to_qp_name(name), {}).get(
            "include-in-capability-statement", False
        ),
        var_name_to_qp_name(name),
    )


@cache
def _load_search_parameters_file() -> Dict[str, Dict[str, Dict[str, Union[str, bool]]]]:
    """
    Load the search parameters JSON file.

    Organize the search parameter file by resource type and return a dict with the data. Initialize
    the search parameters dict with values that aren't present in the JSON file.
    """
    search_parameters = {"Resource": load_extra_search_parameters()}

    bundle = load_search_parameters()

    for entry in bundle["entry"]:
        resource = entry["resource"]

        for resource_type in resource["base"]:
            search_parameters.setdefault(resource_type, {})
            search_parameters[resource_type][resource["name"]] = {
                "type": resource["type"],
                "description": _transform_description(
                    resource["description"], resource_type
                ),
                "uri": entry["fullUrl"],
                "include-in-capability-statement": True,
            }

    return search_parameters


def _transform_description(description: str, resource_type: str) -> str:
    """
    Remove other descriptions in the case where the text contains descriptions for multiple resource
    types, and return only the description for the specified resource type.
    """
    if description.startswith("Multiple Resources:"):
        for description_for_resource_type in description.split("\n"):
            if description_for_resource_type.startswith(f"* [{resource_type}]"):
                _, description = description_for_resource_type.split(": ", maxsplit=1)
                if description.endswith("\r"):
                    return description[:-1]
        else:
            raise AssertionError("Resource type must exist in the description")

    return description



"""Test OpenAPI modifications"""

from typing import Any, Callable, Dict, Mapping, Tuple, cast

import pytest

from .. import FHIRProvider, FHIRStarter, InteractionContext
from ..fhir_specification import FHIR_SEQUENCE
from ..fhirstarter import status
from .config import create_test_client_async
from .resources import Appointment, Bundle, Id, Practitioner


@pytest.fixture(scope="module", autouse=True)
def check_fhir_sequence() -> None:
    """
    Only run tests in this module for FHIR sequence R5.

    This test module runs tests that are specific to R5 to make sure that schema modifications are
    applied properly. There is nothing important about the underlying data, so it can be expected
    that if these tests pass, then schema modifications will work properly for other sequences as
    well.
    """
    if FHIR_SEQUENCE != "R5":
        pytest.skip("blah")


class PractitionerCustom(Practitioner):
    class Config:
        schema_extra = {
            "example": {
                "resourceType": "Practitioner",
                "id": "example",
                "name": [{"family": "Careful", "given": ["Adam"], "prefix": ["Dr"]}],
            }
        }


async def practitioner_read(context: InteractionContext, id_: Id) -> PractitionerCustom:
    raise NotImplementedError


async def practitioner_search_type(context: InteractionContext) -> Bundle:
    raise NotImplementedError


async def appointment_search_type(context: InteractionContext) -> Bundle:
    raise NotImplementedError


@pytest.fixture(scope="module")
def schema() -> Dict[str, Any]:
    client = create_test_client_async(
        ("read", "update", "patch", "delete", "create", "search-type")
    )
    app_ = cast(FHIRStarter, client.app)

    provider = FHIRProvider()

    # Add read and search-type for a resource that uses both the standard model and a model with a
    # custom example
    provider.read(Practitioner)(practitioner_read)
    provider.search_type(PractitionerCustom)(practitioner_search_type)

    # Add a resource that only supports search-type
    provider.search_type(Appointment)(appointment_search_type)

    app_.add_providers(provider)

    return app_.openapi()


def test_inline_search_type_by_post_schemas(schema: Mapping[str, Any]) -> None:
    """Test that search-type by post schemas have been inlined."""
    search_type_paths = (
        (path_name, path)
        for path_name, path in schema["paths"].items()
        if path_name.endswith("_search")
    )
    for path_name, path in search_type_paths:
        keys = path["post"]["requestBody"]["content"][
            "application/x-www-form-urlencoded"
        ]["schema"].keys()
        assert set(keys) == {"properties", "title", "type"}


def test_resource_schemas(schema: Mapping[str, Any]) -> None:
    """Test that all expected schemas are present."""
    schema_names = schema["components"]["schemas"].keys()
    assert set(schema_names) == {
        "Appointment",
        "Bundle",
        "CapabilityStatement",
        "HTTPValidationError",
        "JSONPatchOperation",
        "OperationOutcome",
        "Patient",
        "Practitioner",
        "PractitionerCustom",
        "ValidationError",
    }


def test_content_types(schema: Mapping[str, Any]) -> None:
    """Test that all application/json content types have been changed to application/fhir+json."""
    for path in schema["paths"].values():
        for operation in path.values():
            _, _, interaction_type, *_ = operation["operationId"].split("|")
            if request_body := operation.get("requestBody"):
                content = request_body["content"]
                if interaction_type == "patch":
                    assert "application/json" in content
                elif "application/x-www-form-urlencoded" not in content:
                    assert "application/json" not in content
                    assert "application/fhir+json" in content
            for status_code, response in operation.get("responses", {}).items():
                if status_code == str(status.HTTP_204_NO_CONTENT):
                    assert "content" not in response
                else:
                    content = response["content"]
                    assert "application/json" not in content
                    assert "application/fhir+json" in content


_PATIENT_EXAMPLE_NAMES = (
    "example",
    "pat1",
    "pat2",
    "pat3",
    "pat4",
    "b248b1b2-1686-4b94-9936-37d7a5f94b51",
    "patient-example-sex-and-gender",
    "b3f24cf14-d349-45b3-8134-3d83e9104875",
    "xcda",
    "xds",
    "animal",
    "dicom",
    "ihe-pcd",
    "f001",
    "f201",
    "glossy",
    "proband",
    "genetics-example1",
    "ch-example",
    "newborn",
    "mom",
    "infant-twin-1",
    "infant-twin-2",
    "infant-fetal",
    "infant-mom",
)


@pytest.mark.parametrize(
    argnames="test_id,resource_type,get_examples_func,expected_example_names",
    argvalues=(
        argvalues_me := [
            (
                "capabilities response",
                "CapabilityStatement",
                lambda s: s["paths"]["/metadata"]["get"]["responses"]["200"]["content"][
                    "application/fhir+json"
                ]["examples"],
                (
                    "example",
                    "phr",
                    "capabilitystatement-base",
                    "capabilitystatement-base2",
                    "example-terminology-server",
                    "knowledge-repository",
                    "measure-processor",
                    "messagedefinition",
                ),
            ),
            (
                "Patient read response",
                "Patient",
                lambda s: s["paths"]["/Patient/{id}"]["get"]["responses"]["200"][
                    "content"
                ]["application/fhir+json"]["examples"],
                _PATIENT_EXAMPLE_NAMES,
            ),
            (
                "Patient update request",
                "Patient",
                lambda s: s["paths"]["/Patient/{id}"]["put"]["requestBody"]["content"][
                    "application/fhir+json"
                ]["examples"],
                _PATIENT_EXAMPLE_NAMES,
            ),
            (
                "Patient update response",
                "Patient",
                lambda s: s["paths"]["/Patient/{id}"]["put"]["responses"]["200"][
                    "content"
                ]["application/fhir+json"]["examples"],
                _PATIENT_EXAMPLE_NAMES,
            ),
            (
                "Patient create request",
                "Patient",
                lambda s: s["paths"]["/Patient"]["post"]["requestBody"]["content"][
                    "application/fhir+json"
                ]["examples"],
                _PATIENT_EXAMPLE_NAMES,
            ),
            (
                "Patient create response",
                "Patient",
                lambda s: s["paths"]["/Patient"]["post"]["responses"]["201"]["content"][
                    "application/fhir+json"
                ]["examples"],
                _PATIENT_EXAMPLE_NAMES,
            ),
            (
                "Practitioner read response",
                "Practitioner",
                lambda s: s["paths"]["/Practitioner/{id}"]["get"]["responses"]["200"][
                    "content"
                ]["application/fhir+json"]["examples"],
                (
                    "example",
                    "xcda-author",
                    "3ad0687e-f477-468c-afd5-fcc2bf897809",
                    "f001",
                    "f002",
                    "f003",
                    "f004",
                    "f005",
                    "f201",
                    "f202",
                    "f203",
                    "f204",
                    "f006",
                    "f007",
                    "xcda1",
                    "prac4",
                ),
            ),
        ]
    ),
    ids=[id_ for id_, *_ in argvalues_me],
)
def test_multiple_examples(
    schema: Mapping[str, Any],
    test_id: str,
    resource_type: str,
    get_examples_func: Callable[[Mapping[str, Any]], Dict[str, Any]],
    expected_example_names: Tuple[str, ...],
) -> None:
    """Test that the expected request or response body examples are present."""
    # Test that the expected example names are there
    examples = get_examples_func(schema)
    example_names = tuple(examples.keys())
    assert example_names == expected_example_names

    # Test that the inlined example (the first one) matches the resource type
    first_example_name = example_names[0]
    assert examples[first_example_name]["value"]["resourceType"] == resource_type


@pytest.mark.parametrize(
    argnames="resource_type",
    argvalues=["Appointment", "Patient", "Practitioner"],
    ids=["Appointment", "Patient", "Practitioner"],
)
def test_search_type_examples(
    schema: Mapping[str, Any],
    resource_type: str,
) -> None:
    """
    Test that the resource types in the bundle examples for search-type endpoints are the correct
    resource type.
    """
    example = schema["paths"][f"/{resource_type}"]["get"]["responses"]["200"][
        "content"
    ]["application/fhir+json"]["example"]
    assert example["entry"][0]["resource"]["resourceType"] == resource_type

    example = schema["paths"][f"/{resource_type}/_search"]["post"]["responses"]["200"][
        "content"
    ]["application/fhir+json"]["example"]
    assert example["entry"][0]["resource"]["resourceType"] == resource_type


_OPERATION_OUTCOMES = {
    "400": {
        "resourceType": "OperationOutcome",
        "id": "101",
        "issue": [
            {
                "severity": "error",
                "code": "invalid",
                "details": {"text": "Bad request"},
            }
        ],
    },
    "401": {
        "resourceType": "OperationOutcome",
        "id": "101",
        "issue": [
            {
                "severity": "error",
                "code": "unknown",
                "details": {"text": "Authentication failed"},
            }
        ],
    },
    "403": {
        "resourceType": "OperationOutcome",
        "id": "101",
        "issue": [
            {
                "severity": "error",
                "code": "forbidden",
                "details": {"text": "Authorization failed"},
            }
        ],
    },
    "404": {
        "id": "101",
        "issue": [
            {
                "code": "not-found",
                "details": {"text": "Resource not found"},
                "severity": "error",
            }
        ],
        "resourceType": "OperationOutcome",
    },
    "422": {
        "id": "101",
        "issue": [
            {
                "code": "processing",
                "details": {"text": "Unprocessable entity"},
                "severity": "error",
            }
        ],
        "resourceType": "OperationOutcome",
    },
    "500": {
        "resourceType": "OperationOutcome",
        "id": "101",
        "issue": [
            {
                "severity": "error",
                "code": "exception",
                "details": {"text": "Internal server error"},
            }
        ],
    },
}


@pytest.mark.parametrize(
    argnames="test_id,path,method,expected_status_codes",
    argvalues=(
        argvalues_ee := [
            (
                "Appointment search-type",
                "/Appointment",
                "get",
                ("400", "401", "403", "500"),
            ),
            (
                "Appointment search-type by post",
                "/Appointment/_search",
                "post",
                ("400", "401", "403", "500"),
            ),
            ("Patient read", "/Patient/{id}", "get", ("401", "403", "404", "500")),
            (
                "Patient update",
                "/Patient/{id}",
                "put",
                ("400", "401", "403", "404", "422", "500"),
            ),
            ("Patient create", "/Patient", "post", ("400", "401", "403", "422", "500")),
            ("Patient search-type", "/Patient", "get", ("400", "401", "403", "500")),
            (
                "Patient search-type by post",
                "/Patient/_search",
                "post",
                ("400", "401", "403", "500"),
            ),
            (
                "Practitioner read",
                "/Practitioner/{id}",
                "get",
                ("401", "403", "404", "500"),
            ),
            (
                "Practitioner search-type",
                "/Practitioner",
                "get",
                ("400", "401", "403", "500"),
            ),
            (
                "Practitioner search-type by post",
                "/Practitioner/_search",
                "post",
                ("400", "401", "403", "500"),
            ),
        ]
    ),
    ids=[id_ for id_, *_ in argvalues_ee],
)
def test_error_examples(
    test_id: str,
    schema: Mapping[str, Any],
    path: str,
    method: str,
    expected_status_codes: Tuple[str, ...],
) -> None:
    """Test that the expected error response body examples are present."""
    responses = schema["paths"][path][method]["responses"]

    # Test that all of the expected error status codes are present
    status_codes = set(responses.keys())
    status_codes.discard("200")
    status_codes.discard("201")
    assert status_codes == set(expected_status_codes)

    # Test that all the error examples are present
    for status_code in expected_status_codes:
        assert (
            responses[status_code]["content"]["application/fhir+json"]["example"]
            == _OPERATION_OUTCOMES[status_code]
        )



"""Script to populate the FHIR specification data for each of the sequences"""

import concurrent
import json
import sys
import zipfile
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Any, Dict, Literal, Union
from uuid import uuid4

import requests
from bs4 import BeautifulSoup, Tag

ADVERSE_EVENT_R5_EXAMPLE = {
    "summary": "Example of adverseevent",
    "description": "Example of adverseevent",
    "value": {
        "resourceType": "AdverseEvent",
        "id": "example",
        "text": {
            "status": "generated",
            "div": '<div xmlns="http://www.w3.org/1999/xhtml"><p><b>Generated Narrative</b></p><div style="display: inline-block; background-color: #d9e0e7; padding: 6px; margin: 4px; border: 1px solid #8da1b4; border-radius: 5px; line-height: 60%"><p style="margin-bottom: 0px">Resource &quot;example&quot; </p></div><p><b>identifier</b>: id: 49476534</p><p><b>actuality</b>: actual</p><p><b>category</b>: Medication Mishap <span style="background: LightGoldenRodYellow; margin: 4px; border: 1px solid khaki"> (<a href="codesystem-adverse-event-category.html">AdverseEventCategory</a>#medication-mishap)</span></p><p><b>event</b>: This was a mild rash on the left forearm <span style="background: LightGoldenRodYellow; margin: 4px; border: 1px solid khaki"> (<a href="https://browser.ihtsdotools.org/">SNOMED CT</a>#304386008 &quot;O/E - itchy rash&quot;)</span></p><p><b>subject</b>: <a href="patient-example.html">Patient/example</a> &quot;Peter CHALMERS&quot;</p><p><b>date</b>: 2017-01-29T12:34:56Z</p><p><b>seriousness</b>: Non-serious <span style="background: LightGoldenRodYellow; margin: 4px; border: 1px solid khaki"> (<a href="codesystem-adverse-event-seriousness.html">AdverseEventSeriousness</a>#Non-serious)</span></p><p><b>recorder</b>: <a href="practitioner-example.html">Practitioner/example</a> &quot;Adam CAREFUL&quot;</p><h3>SuspectEntities</h3><table class="grid"><tr><td>-</td><td><b>Instance</b></td></tr><tr><td>*</td><td><a href="todo.html">Medication/example</a></td></tr></table></div>',
        },
        "identifier": [
            {"system": "http://acme.com/ids/patients/risks", "value": "49476534"}
        ],
        "status": "completed",
        "actuality": "actual",
        "category": [
            {
                "coding": [
                    {
                        "system": "http://terminology.hl7.org/CodeSystem/adverse-event-category",
                        "code": "medication-mishap",
                        "display": "Medication Mishap",
                    }
                ]
            }
        ],
        "code": {
            "coding": [
                {
                    "system": "http://snomed.info/sct",
                    "code": "304386008",
                    "display": "O/E - itchy rash",
                }
            ],
            "text": "This was a mild rash on the left forearm",
        },
        "subject": {"reference": "Patient/example"},
        "occurrenceDateTime": "2017-01-29T12:34:56+00:00",
        "seriousness": {
            "coding": [
                {
                    "system": "http://terminology.hl7.org/CodeSystem/adverse-event-seriousness",
                    "code": "non-serious",
                    "display": "Non-serious",
                }
            ]
        },
        "recorder": {"reference": "Practitioner/example"},
        "suspectEntity": [{"instanceReference": {"reference": "Medication/example"}}],
    },
}


def main() -> None:
    fhir_dir = Path(sys.argv[1])

    print("")

    for sequence in ("STU3", "R4", "R4B", "R5"):
        print(f"Creating sequence data for {sequence}...")

        # Download the search parameters file
        response = requests.get(
            f"https://hl7.org/fhir/{sequence}/search-parameters.json"
        )

        # Make a zip file with the search parameters file
        with zipfile.ZipFile(
            fhir_dir / "sequences" / sequence / "search-parameters.zip",
            "w",
            compression=zipfile.ZIP_DEFLATED,
            compresslevel=9,
        ) as file_:
            file_.writestr("search-parameters.json", response.content)

        # Load the list of resources
        with open(fhir_dir / "sequences" / sequence / "resource_types.json") as file_:
            resource_types = json.load(file_)

        # Get the examples for all resource types
        examples = {}
        with ThreadPoolExecutor() as executor:
            future_to_resource_type = {
                executor.submit(get_examples, sequence, resource_type): resource_type
                for resource_type in resource_types
            }
            for future in concurrent.futures.as_completed(future_to_resource_type):
                resource_type = future_to_resource_type[future]
                examples[resource_type] = future.result()

        # Create the examples zip file
        with zipfile.ZipFile(
            fhir_dir / "sequences" / sequence / "examples.zip",
            "w",
            compression=zipfile.ZIP_DEFLATED,
            compresslevel=9,
        ) as file_:
            for resource_type in examples.keys():
                file_.writestr(
                    f"{resource_type.lower()}.json",
                    json.dumps(
                        examples[resource_type], indent=4, separators=(", ", ": ")
                    ),
                )

    print("\nDone")


def get_examples(sequence: str, resource_type: str) -> Dict[str, Any]:
    if sequence == "R5" and resource_type == "AdverseEvent":
        return {
            "example": {
                "summary": "Example of adverseevent",
                "description": "Example of adverseevent",
                "value": ADVERSE_EVENT_R5_EXAMPLE,
            }
        }

    if resource_type == "StructureDefinition":
        # StructureDefinition is materially different than other resource types
        examples = _get_structuredefinition_examples(sequence)
    else:
        # Download the examples page for the resource type
        response = requests.get(
            f"https://hl7.org/fhir/{sequence}/{resource_type.lower()}-examples.html"
        )
        if response.status_code != requests.codes.ok:
            raise RuntimeError(f"Failed to get list of examples for {resource_type}")

        # Extract the description, identifier, and JSON file URL for each example
        soup = BeautifulSoup(response.text, features="html.parser")
        examples = _get_examples(sequence, resource_type, soup.find_all("table")[-1])

    # Inline the first example
    first_example = examples[next(iter(examples.keys()))]
    response = requests.get(first_example["externalValue"])
    if response.status_code != requests.codes.ok:
        raise RuntimeError(f"Failed to download example for {resource_type}")
    first_example["value"] = response.json()
    del first_example["externalValue"]

    return examples


def _get_structuredefinition_examples(sequence: str) -> Dict[str, Any]:
    resource_type = "StructureDefinition"

    # Download the examples page for the resource type
    response = requests.get(
        f"https://hl7.org/fhir/{sequence}/{resource_type.lower()}-examples.html"
    )
    if response.status_code != requests.codes.ok:
        raise RuntimeError(f"Failed to get list of examples for {resource_type}")

    # Extract the description, identifier, and JSON file URL for each example
    soup = BeautifulSoup(response.text, features="html.parser")
    examples = _get_examples(
        sequence,
        resource_type,
        soup.find("div", attrs={"id": "tabs-1"}),
        description_prefix="Base Type",
        id_method="random",
    )
    examples.update(
        _get_examples(
            sequence,
            resource_type,
            soup.find("div", attrs={"id": "tabs-2"}),
            description_prefix="Resource",
            id_method="random",
        )
    )
    examples.update(
        _get_examples(
            sequence,
            resource_type,
            soup.find("div", attrs={"id": "tabs-3"}),
            description_prefix="Constraint",
            id_method="random",
        )
    )
    examples.update(
        _get_examples(
            sequence,
            resource_type,
            soup.find("div", attrs={"id": "tabs-4"}),
            description_prefix="Extension",
            id_method="description",
        )
    )
    examples.update(
        _get_examples(
            sequence,
            resource_type,
            soup.find("div", attrs={"id": "tabs-5"}),
            description_prefix="Example",
            id_method="standard",
        )
    )

    return examples


def _get_examples(
    sequence: str,
    resource_type: str,
    examples_table: Union[Tag, None],
    description_prefix: str = "",
    id_method: Literal["standard", "random", "description"] = "standard",
) -> Dict[str, Any]:
    examples: Dict[str, Any] = {}

    if not examples_table:
        return examples

    # Iterate over the examples and extract the description, identifier, and JSON file URL
    for example in examples_table.find_all("tr"):
        # Skip any rows that don't have any URLs in them
        if len(example.find_all("a")) == 0:
            continue

        cells = example.find_all("td")

        # Get the description and ID based on the specified method
        if id_method == "standard":
            description = cells[0].text
            id_ = cells[1].text
        elif id_method == "random":
            description = cells[0].text
            id_ = uuid4().hex
        elif id_method == "description":
            a = cells[0].find("a")
            description = id_ = a.text
        else:
            raise AssertionError(
                f"Unable to get description and ID for {resource_type} example"
            )

        # Add a description prefix if one is specified
        if description_prefix:
            description = f"{description_prefix}: {description}"

        # Find the JSON filename
        for a in example.find_all("a"):
            if a.text.lower() == "json":
                filename = a.attrs["href"]
                if filename.endswith(".html"):
                    filename = filename[:-5]
                break
        else:
            raise AssertionError(
                f"Unable to find JSON filename for {resource_type} example"
            )

        # Add the example in OpenAPI format
        examples[id_] = {
            "summary": description,
            "description": description,
            "externalValue": f"https://hl7.org/fhir/{sequence}/{filename}",
        }

    return examples


if __name__ == "__main__":
    main()



"""FHIRStarter test configuration"""

from copy import deepcopy
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, Tuple, Union, cast

import jsonpatch

from ..exceptions import FHIRResourceNotFoundError
from ..fhirstarter import FHIRStarter
from ..interactions import InteractionContext
from ..json_patch import JSONPatch, convert_json_patch
from ..providers import FHIRProvider
from ..resources import Bundle, Id
from ..testclient import TestClient
from .resources import HumanName, Patient
from .utils import generate_fhir_resource_id

# In-memory "database" used to simulate persistence of created FHIR resources
DATABASE: Dict[str, Patient] = {}

_VALID_TOKEN = "valid"
_INVALID_TOKEN = "invalid"


async def patient_read_async(context: InteractionContext, id_: Id) -> Patient:
    """Patient read FHIR interaction."""
    return patient_read(context, id_)


def patient_read(_: InteractionContext, id_: Id) -> Patient:
    """Patient read FHIR interaction."""
    patient = DATABASE.get(id_)
    if not patient:
        raise FHIRResourceNotFoundError

    return patient


async def patient_update_async(
    context: InteractionContext, id_: Id, resource: Patient
) -> Id:
    """Patient update FHIR interaction."""
    return patient_update(context, id_, resource)


def patient_update(_: InteractionContext, id_: Id, resource: Patient) -> Id:
    """Patient update FHIR interaction."""
    if id_ not in DATABASE:
        raise FHIRResourceNotFoundError

    patient = deepcopy(resource)
    DATABASE[id_] = patient

    return Id(patient.id)


async def patient_patch_async(
    context: InteractionContext, id_: Id, json_patch: JSONPatch
) -> Id:
    """Patient patch FHIR interaction."""
    return patient_patch(context, id_, json_patch)


def patient_patch(_: InteractionContext, id_: Id, json_patch: JSONPatch) -> Id:
    """Patient patch FHIR interaction."""
    if id_ not in DATABASE:
        raise FHIRResourceNotFoundError

    patient = DATABASE[id_].dict()

    jsonpatch.apply_patch(
        patient, jsonpatch.JsonPatch(convert_json_patch(json_patch)), in_place=True
    )

    DATABASE[id_] = Patient(**patient)

    return Id(id_)


async def patient_delete_async(context: InteractionContext, id_: Id) -> None:
    """Patient delete FHIR interaction."""
    return patient_delete(context, id_)


def patient_delete(_: InteractionContext, id_: Id) -> None:
    """Patient delete FHIR interaction."""
    if id_ not in DATABASE:
        return

    del DATABASE[id_]

    return None


async def patient_create_async(context: InteractionContext, resource: Patient) -> Id:
    """Patient create FHIR interaction."""
    return patient_create(context, resource)


def patient_create(_: InteractionContext, resource: Patient) -> Id:
    """Patient create FHIR interaction."""
    patient = deepcopy(resource)
    patient.id = generate_fhir_resource_id()
    DATABASE[patient.id] = patient

    return Id(patient.id)


async def patient_search_type_async(
    context: InteractionContext,
    family: Union[str, None],
    general_practitioner: Union[str, None],
    nickname: Union[str, None],
    _last_updated: Union[str, None],
) -> Bundle:
    """Patient search-type FHIR interaction."""
    return patient_search_type(
        context, family, general_practitioner, nickname, _last_updated
    )


def patient_search_type(
    _: InteractionContext,
    family: Union[str, None],
    general_practitioner: Union[str, None],
    nickname: Union[str, None],
    _last_updated: Union[str, None],
) -> Bundle:
    """Patient search-type FHIR interaction."""
    patients = []
    for patient in DATABASE.values():
        for name in patient.name:
            if cast(HumanName, name).family == family:
                patients.append(patient)

    bundle = Bundle(
        **{
            "type": "searchset",
            "total": len(patients),
            "entry": [{"resource": patient.dict()} for patient in patients],
        }
    )

    return bundle


def app(provider: FHIRProvider) -> TestClient:
    """Create a FHIRStarter app, add the provider, reset the database, and return a TestClient."""
    config_file_contents = """
[search-parameters.Patient.nickname]
type = "string"
description = "Nickname"
uri = "https://hostname/nickname"
include-in-capability-statement = true
    """

    with TemporaryDirectory() as path:
        config_file = Path(path) / "config.toml"
        with open(config_file, "w") as file_:
            file_.write(config_file_contents)

        app_ = FHIRStarter(config_file=config_file)

    app_.add_providers(provider)

    DATABASE.clear()

    return TestClient(app_)


def create_test_client_async(interactions: Tuple[str, ...]) -> TestClient:
    """Given a list of interactions, create an app with async handlers and return a test client."""
    provider = FHIRProvider()

    for interaction in interactions:
        if interaction == "read":
            provider.read(Patient)(patient_read_async)
        elif interaction == "update":
            provider.update(Patient)(patient_update_async)
        elif interaction == "patch":
            provider.patch(Patient)(patient_patch_async)
        elif interaction == "delete":
            provider.delete(Patient)(patient_delete_async)
        elif interaction == "create":
            provider.create(Patient)(patient_create_async)
        elif interaction == "search-type":
            provider.search_type(Patient)(patient_search_type_async)

    return app(provider)


def create_test_client(interactions: Tuple[str, ...]) -> TestClient:
    """Given a list of interactions, create an app and return a test client."""
    provider = FHIRProvider()

    for interaction in interactions:
        if interaction == "read":
            provider.read(Patient)(patient_read)
        elif interaction == "update":
            provider.update(Patient)(patient_update)
        elif interaction == "patch":
            provider.patch(Patient)(patient_patch)
        elif interaction == "delete":
            provider.delete(Patient)(patient_delete)
        elif interaction == "create":
            provider.create(Patient)(patient_create)
        elif interaction == "search-type":
            provider.search_type(Patient)(patient_search_type)

    return app(provider)



"""FHIRStarter class, exception handlers, and middleware."""

import itertools
import logging

try:
    import tomllib
except ImportError:
    import tomli as tomllib

import datetime
from collections import defaultdict
from io import IOBase
from os import PathLike
from typing import (
    Any,
    Callable,
    Coroutine,
    DefaultDict,
    Dict,
    List,
    MutableMapping,
    Union,
    cast,
)
from urllib.parse import parse_qs, urlencode

from fastapi import FastAPI, HTTPException, Request, Response, status
from fastapi.exceptions import RequestValidationError
from pydantic.error_wrappers import display_errors

from .exceptions import FHIRException
from .fhir_specification import FHIR_SEQUENCE, FHIR_VERSION
from .functions import (
    FORMAT_QP,
    PRETTY_QP,
    make_create_function,
    make_delete_function,
    make_patch_function,
    make_read_function,
    make_search_type_function,
    make_update_function,
)
from .interactions import ResourceType, TypeInteraction
from .openapi import adjust_schema
from .providers import FHIRProvider
from .resources import CapabilityStatement, OperationOutcome
from .search_parameters import (
    SearchParameters,
    search_parameter_sort_key,
    supported_search_parameters,
    var_name_to_qp_name,
)
from .utils import (
    FormatParameters,
    create_route_args,
    delete_route_args,
    format_response,
    make_operation_outcome,
    parse_fhir_request,
    patch_route_args,
    read_route_args,
    search_type_route_args,
    update_route_args,
)

# Suppress warnings from base fhir.resources class
logging.getLogger("fhir.resources.core.fhirabstractmodel").setLevel(logging.WARNING + 1)

CapabilityStatementModifier = Callable[
    [MutableMapping[str, Any], Request, Response], MutableMapping[str, Any]
]

_INTERACTION_ORDER = {
    "read": 1,
    "vread": 2,
    "update": 3,
    "patch": 4,
    "delete": 5,
    "history-instance": 6,
    "history-type": 7,
    "create": 8,
    "search-type": 9,
}


class FHIRStarter(FastAPI):
    """
    FHIRStarter class.

    Handles collection of FHIR providers, creation of API routes, middleware, exception handling,
    and capability statement requests.
    """

    def __init__(
        self,
        *,
        config_file: Union[str, PathLike, IOBase, None] = None,
        title: str = "FHIRStarter",
        **kwargs: Any,
    ) -> None:
        """
        On app creation, the following occurs:
        * Custom search parameters are loaded
        * Static routes are created (e.g. the capability statement route)
        * Middleware is added (e.g. content-type header handling)
        * Exception handling is added
        """
        super().__init__(title=title, **kwargs)

        if config_file:
            try:
                cast(IOBase, config_file).seek(0)
                config = tomllib.load(config_file)
            except AttributeError:
                with open(config_file, "rb") as file_:
                    config = tomllib.load(file_)

            self._search_parameters = SearchParameters(config.get("search-parameters"))
        else:
            self._search_parameters = SearchParameters()

        self._capabilities: DefaultDict[str, Dict[str, TypeInteraction]] = defaultdict(
            dict
        )
        self._created = (
            datetime.datetime.now().astimezone(datetime.timezone.utc).isoformat()
        )

        self.set_capability_statement_modifier(lambda c, _, __: c)

        self._add_capabilities_route()

        self.middleware("http")(_transform_search_type_post_request)
        self.middleware("http")(_transform_null_response_body)
        self.middleware("http")(_set_content_type_header)

        async def default_exception_callback(
            _: Request, response: Response, __: Exception
        ) -> Response:
            return response

        self._exception_callback: Callable[
            [Request, Response, Exception], Coroutine[None, None, Response]
        ] = default_exception_callback

        self.add_exception_handler(
            RequestValidationError, self.validation_exception_handler
        )
        self.add_exception_handler(HTTPException, self.http_exception_handler)
        self.add_exception_handler(FHIRException, self.fhir_exception_handler)
        self.add_exception_handler(Exception, self.general_exception_handler)

    def add_providers(self, *providers: FHIRProvider) -> None:
        """
        Add all FHIR interactions from all provided FHIRProviders.

        Iterate over the interactions from the providers, record the capabilities (based on the
        resource type and interaction type), and add the API route for the defined interaction.
        """
        provider_interactions = itertools.chain.from_iterable(
            provider.interactions for provider in providers
        )

        for interaction in sorted(
            sorted(
                provider_interactions,
                key=lambda i: cast(str, i.resource_type.get_resource_type()),
            ),
            key=lambda i: _INTERACTION_ORDER[i.label()],
        ):
            resource_type = interaction.resource_type.get_resource_type()
            label = interaction.label()
            assert (
                resource_type not in self._capabilities
                or label not in self._capabilities[resource_type]
            ), f"FHIR {label} interaction for {resource_type} can only be supplied once"

            self._capabilities[resource_type][label] = interaction
            self._add_route(interaction)

    def set_capability_statement_modifier(
        self, modifier: CapabilityStatementModifier
    ) -> None:
        """
        Set a user-provided callable that will make adjustments to the automatically-generated
        capability statement.

        The user-provided callable must take a mutable mapping, which will be the capability
        statement, and return a mutable mapping, which will be the modified version of the
        capability statement.

        This method enables any desired change to be made to the capability statement, such as
        filling in fields that are not automatically generated, or adding extensions.

        All modifications made to the capability statement must conform to the specification of the
        FHIR CapabilityStatement resource, or server startup will fail.
        """
        self._capability_statement_modifier = modifier

    def set_exception_callback(
        self,
        callback: Callable[
            [Request, Response, Exception], Coroutine[None, None, Response]
        ],
    ) -> None:
        """
        Set a user-provided callback function that will run whenever any type of exception occurs.
        This configuration option is useful for injecting additional exception handling behavior,
        such as exception logging.
        """
        self._exception_callback = callback

    async def validation_exception_handler(
        self, request: Request, exception: RequestValidationError
    ) -> Response:
        """
        Validation exception handler that overrides the default FastAPI validation exception
        handler.

        Creates an operation outcome by destructuring the RequestValidationError and mapping the
        values to the correct places in the OperationOutcome.
        """
        operation_outcome = OperationOutcome(
            **{
                "issue": [
                    {
                        "severity": "error",
                        "code": _pydantic_error_to_fhir_issue_type(error["type"]),
                        "details": {
                            "text": display_errors([error]).replace("\n ", " —")
                        },
                    }
                    for error in exception.errors()
                ]
            }
        )

        response = format_response(
            resource=operation_outcome,
            status_code=status.HTTP_400_BAD_REQUEST,
            format_parameters=FormatParameters.from_request(
                request, raise_exception=False
            ),
        )

        return await self._exception_callback(request, response, exception)

    async def http_exception_handler(
        self, request: Request, exception: HTTPException
    ) -> Response:
        """
        HTTP exception handler that overrides the default FastAPI HTTP exception handler.

        This exception handler exists primarily to convert an HTTP exception into an
        OperationOutcome.
        """
        response = _exception_response(
            request=request,
            severity="error",
            code="processing",
            details_text=exception.detail,
            status_code=exception.status_code,
        )

        return await self._exception_callback(request, response, exception)

    async def fhir_exception_handler(
        self, request: Request, exception: FHIRException
    ) -> Response:
        """
        General exception handler to catch all other FHIRExceptions. Returns an OperationOutcome.

        Set the request on the exception so that the exception has more context with which to form
        an OperationOutcome.
        """
        exception.set_request(request)

        response = format_response(
            resource=exception.operation_outcome(),
            status_code=exception.status_code,
            format_parameters=FormatParameters.from_request(
                request, raise_exception=False
            ),
        )

        return await self._exception_callback(request, response, exception)

    async def general_exception_handler(
        self, request: Request, exception: Exception
    ) -> Response:
        """
        General exception handler to catch server framework errors. Returns an OperationOutcome.
        """
        response = _exception_response(
            request=request,
            severity="error",
            code="exception",
            details_text=str(exception),
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
        )

        return await self._exception_callback(request, response, exception)

    def capability_statement(
        self, request: Request, response: Response
    ) -> CapabilityStatement:
        """
        Generate the capability statement for the instance based on the FHIR interactions provided.

        In addition to declaring the interactions, the supported search parameters are also
        declared.
        """
        resources = []
        for resource_type, interactions in sorted(self._capabilities.items()):
            search_parameter_metadata = self._search_parameters.get_metadata(
                resource_type
            )

            resource = {
                "type": resource_type,
                "interaction": [
                    {"code": label}
                    for label in sorted(
                        interactions.keys(), key=lambda l: _INTERACTION_ORDER[l]
                    )
                ],
            }
            if search_type_interaction := interactions.get("search-type"):
                supported_search_parameters_ = []
                for search_parameter in supported_search_parameters(
                    search_type_interaction.handler
                ):
                    search_parameter_name = var_name_to_qp_name(search_parameter.name)
                    metadata = search_parameter_metadata[search_parameter_name]
                    if metadata["include-in-capability-statement"]:
                        supported_search_parameters_.append(
                            {
                                "name": search_parameter_name,
                                "definition": metadata["uri"],
                                "type": metadata["type"],
                                "documentation": metadata["description"],
                            }
                        )
                resource["searchParam"] = sorted(
                    supported_search_parameters_,
                    key=lambda p: search_parameter_sort_key(
                        cast(Dict[str, str], p)["name"], search_parameter_metadata
                    ),
                )
            resources.append(resource)

        capability_statement = {
            "status": "active",
            "date": self._created,
            "kind": "instance",
            "fhirVersion": FHIR_VERSION,
            "acceptUnknown": "no",
            "format": ["json"],
            "rest": [
                {
                    "mode": "server",
                    "resource": resources,
                }
            ],
        }

        if FHIR_SEQUENCE != "STU3":
            del capability_statement["acceptUnknown"]

        return CapabilityStatement(
            **self._capability_statement_modifier(
                capability_statement, request, response
            )
        )

    def openapi(self) -> Dict[str, Any]:
        """Adjust the OpenAPI schema to make it more FHIR-friendly."""
        if self.openapi_schema:
            return self.openapi_schema

        openapi_schema = super().openapi()
        adjust_schema(openapi_schema)

        return openapi_schema

    def _add_capabilities_route(self) -> None:
        """Add the /metadata route, which supplies the capability statement for the instance."""

        def capability_statement_handler(
            request: Request,
            response: Response,
            _format: str = FORMAT_QP,
            _pretty: str = PRETTY_QP,
        ) -> Union[CapabilityStatement, Response]:
            return format_response(
                resource=self.capability_statement(request, response),
                response=response,
                format_parameters=FormatParameters.from_request(request),
            )

        self.get(
            "/metadata",
            response_model=CapabilityStatement,
            status_code=status.HTTP_200_OK,
            tags=["System"],
            summary="capabilities",
            description="The capabilities interaction retrieves the information about a server's "
            "capabilities - which portions of the FHIR specification it supports.",
            operation_id="fhirstarter|system|capabilities|get",
            response_model_exclude_none=True,
        )(capability_statement_handler)

    def _add_route(self, interaction: TypeInteraction[ResourceType]) -> None:
        """
        Add a route based on the FHIR interaction type.

        FHIR search-type routes must support both GET and POST, so two routes are added for
        search-type interactions.
        """
        if interaction.label() == "read":
            self.get(**read_route_args(interaction))(make_read_function(interaction))
        elif interaction.label() == "update":
            self.put(**update_route_args(interaction))(
                make_update_function(interaction)
            )
        elif interaction.label() == "patch":
            self.patch(**patch_route_args(interaction))(
                make_patch_function(interaction)
            )
        elif interaction.label() == "delete":
            self.delete(**delete_route_args(interaction))(
                make_delete_function(interaction)
            )
        elif interaction.label() == "create":
            self.post(**create_route_args(interaction))(
                make_create_function(interaction)
            )
        elif interaction.label() == "search-type":
            search_parameter_metadata = self._search_parameters.get_metadata(
                interaction.resource_type.get_resource_type()
            )
            self.get(**search_type_route_args(interaction, post=False))(
                make_search_type_function(
                    interaction,
                    search_parameter_metadata=search_parameter_metadata,
                    post=False,
                )
            )
            self.post(**search_type_route_args(interaction, post=True))(
                make_search_type_function(
                    interaction,
                    search_parameter_metadata=search_parameter_metadata,
                    post=True,
                )
            )


async def _transform_search_type_post_request(
    request: Request, call_next: Callable[[Request], Coroutine[None, None, Response]]
) -> Response:
    """
    Middleware that transforms a search POST request into a search GET request.

    This is needed for a few reasons, and mainly to simplify how searches are handled later down the
    line. Due to this middleware, all search requests will arrive in the handlers as GET requests
    with query strings that have been merged with the URL-encoded parameter string in the body.

    There is an obscure requirement in the FHIR specification stipulating that for search POST
    requests, both query string parameters and parameters in the body are to be considered when
    calculating search results. This is difficult to achieve in FastAPI due to how the body stream
    is consumed when it parses the body to pass the values down to the handlers. Catching the
    request here allows for the body parameters to be merged with the query string parameters.
    """
    interaction_info = parse_fhir_request(request)

    if (
        interaction_info.interaction_type == "search-type"
        and request.method == "POST"
        and request.headers.get("Content-Type") == "application/x-www-form-urlencoded"
    ):
        scope = request.scope
        scope["method"] = "GET"
        if scope["path"].endswith("/_search"):
            scope["path"] = scope["path"][:-8]
        if scope["raw_path"][-8:] == "/_search":
            scope["raw_path"] = scope["raw_path"][:-8]
        scope["query_string"] = await _merge_parameter_strings(request)
        scope["headers"] = [
            (name, value)
            for name, value in scope["headers"]
            if name.lower() not in {b"content-length", b"content-type"}
        ]

        return await call_next(Request(scope, request.receive))

    return await call_next(request)


async def _transform_null_response_body(
    request: Request, call_next: Callable[[Request], Coroutine[None, None, Response]]
) -> Response:
    """
    Middleware that cleans up the response object when the response does not contain a response
    body.

    Update, patch, and create interactions are not required to return a response body. In this
    scenario, FastAPI for some reason returns a response with a body containing the string "null",
    rather than just an empty body. This middleware detects that scenario and cleans up the response
    body.
    """
    response = await call_next(request)

    # Clean up the content headers for delete interactions
    interaction_info = parse_fhir_request(request)
    if (
        interaction_info.interaction_type == "delete"
        and "Content-Length" not in response.headers
    ):
        response.headers["Content-Length"] = "0"
        if "Content-Type" in response.headers:
            del response.headers["Content-Type"]

    # This condition is probably too broad, but given that all FHIR responses should either have a
    # body that is a resource (that must be more than 4 bytes by definition) or an empty response
    # body, it is safe to assume that a 4-byte response contains the string "null" and needs to be
    # transformed.
    if (
        "application/fhir+json" in response.headers.getlist("Content-Type")
        and response.headers.get("Content-Length") == "4"
    ):
        response.headers["Content-Length"] = "0"
        del response.headers["Content-Type"]
        response = Response(status_code=response.status_code, headers=response.headers)

    return response


async def _merge_parameter_strings(request: Request) -> bytes:
    """
    Merge the query string and the parameter string in the body into a single parameter string.

    If there is a header that specifies the requested format, then ignore the _format parameter(s)
    in the parameter strings.
    """
    merged: DefaultDict[bytes, List[bytes]] = defaultdict(list)

    format_ = FormatParameters.format_from_accept_header(request)
    if format_:
        merged[b"_format"] = [format_.encode()]

    for query_string in (await request.body(), request.scope["query_string"]):
        for name, values in parse_qs(query_string).items():
            if format_ and name == "_format":
                continue
            merged[name].extend(values)

    return urlencode(merged, doseq=True).encode()


async def _set_content_type_header(
    request: Request, call_next: Callable[[Request], Coroutine[None, None, Response]]
) -> Response:
    """
    Middleware that changes the content type header to "application/fhir+json".

    For FHIR responses, there will be two content type headers in the response. One will be
    "application/json" (added by FastAPI), and one will be "application/fhir+json" (added by
    FHIRStarter). This middleware removes the "application/json" header.
    """
    response: Response = await call_next(request)

    if "application/fhir+json" in response.headers.getlist("Content-Type"):
        response.headers["Content-Type"] = "application/fhir+json"

    return response


def _pydantic_error_to_fhir_issue_type(error: str) -> str:
    """Return a FHIR issue type code mapped from a Pydantic error code."""
    error_type, *rest = error.split(".")
    error_code = rest[0] if rest else None

    if error_type == "json_invalid":
        return "structure"
    elif error_type == "type_error":
        return "value"
    elif error_type == "value_error":
        if error_code == "extra":
            return "structure"
        elif error_code == "missing":
            return "required"
        else:
            return "value"
    else:
        return "invalid"


def _exception_response(
    request: Request, severity: str, code: str, details_text: str, status_code: int
) -> Response:
    """Create a JSONResponse with an OperationOutcome and an HTTP status code."""
    operation_outcome = make_operation_outcome(
        severity=severity, code=code, details_text=details_text or "Exception"
    )

    return format_response(
        resource=operation_outcome,
        status_code=status_code,
        format_parameters=FormatParameters.from_request(request, raise_exception=False),
    )



"""
Dynamic function creation for FHIR interactions.

The callables passed to FastAPI by FHIRStarter are created using functional programming techniques.
Interactions other than search are fairly straightforward -- these functions simply call a
developer-provided handler, perform some FHIR-related processing, and return the result up the
chain.

The search use case is slightly more complicated. Because each FHIR resource type has a different
set of search parameters (i.e. query parameters), the approach used for the read, update, and create
use cases to create the callables does not work. Instead, a function that takes arbitrary kwargs is
created, and the function signature is modified afterward to provide the information needed by
FastAPI for documentation purposes: variable name, annotation, and default value.

These functions can create FastAPI path operation functions for both async and non-async handlers,
though not elegantly. Unfortunately the code has to be duplicated, and maintainers will need to be
cognizant that changes made to these generated path operation functions must be applied in two
places.
"""

import keyword
from inspect import Parameter, iscoroutinefunction, signature
from typing import Callable, Coroutine, Dict, List, Tuple, Union, cast

from fastapi import Form, Path, Query, Request, Response

from .exceptions import FHIRBadRequestError
from .interactions import (
    CreateInteractionHandler,
    DeleteInteractionHandler,
    InteractionContext,
    PatchInteractionHandler,
    ReadInteractionHandler,
    ResourceType,
    SearchTypeInteractionHandler,
    TypeInteraction,
    UpdateInteractionHandler,
)
from .json_patch import JSONPatch
from .resources import Bundle, Id, Resource
from .search_parameters import (
    search_parameter_sort_key,
    supported_search_parameters,
    var_name_to_qp_name,
)
from .utils import FormatParameters, format_response

_FORMAT_PARAMETER_DESCRIPTION = (
    "Override the HTTP content negotiation to specify JSON or XML response format"
)
_PRETTY_PARAMETER_DESCRIPTION = (
    "Ask for a pretty printed response for human convenience"
)

FORMAT_QP = Query(None, description=_FORMAT_PARAMETER_DESCRIPTION)
PRETTY_QP = Query(None, description=_PRETTY_PARAMETER_DESCRIPTION)


# Note: I'm not currently aware of a better way do support both async and non-async handlers than
#       branching based on the result of the iscoroutinefunction function and duplicating most of
#       the code. If there is a better way to do this, I will refactor later.


def make_read_function(
    interaction: TypeInteraction[ResourceType],
) -> Callable[
    [Request, Response, Id, str, str],
    Union[Coroutine[None, None, Union[ResourceType, Response]], ResourceType, Response],
]:
    """Make a function suitable for creation of a FHIR read API route."""

    if iscoroutinefunction(interaction.handler):

        async def read_async(
            request: Request,
            response: Response,
            id_: Id = Path(
                alias="id",
                description=Resource.schema()["properties"]["id"]["title"],
            ),
            _format: str = FORMAT_QP,
            _pretty: str = PRETTY_QP,
        ) -> Union[ResourceType, Response]:
            """Function for read interaction."""
            handler = cast(ReadInteractionHandler[ResourceType], interaction.handler)
            result_resource = await handler(InteractionContext(request, response), id_)  # type: ignore[call-arg]

            return format_response(
                resource=result_resource,
                response=response,
                format_parameters=FormatParameters.from_request(request),
            )

        return read_async

    else:

        def read(
            request: Request,
            response: Response,
            id_: Id = Path(
                alias="id",
                description=Resource.schema()["properties"]["id"]["title"],
            ),
            _format: str = FORMAT_QP,
            _pretty: str = PRETTY_QP,
        ) -> Union[ResourceType, Response]:
            """Function for read interaction."""
            handler = cast(ReadInteractionHandler[ResourceType], interaction.handler)
            result_resource = handler(InteractionContext(request, response), id_)  # type: ignore[call-arg]

            return format_response(
                resource=result_resource,
                response=response,
                format_parameters=FormatParameters.from_request(request),
            )

        return read


def make_update_function(
    interaction: TypeInteraction[ResourceType],
) -> Callable[
    [Request, Response, ResourceType, Id, str, str],
    Union[Coroutine[None, None, Union[ResourceType, Response]], ResourceType, Response],
]:
    """Make a function suitable for creation of a FHIR update API route."""
    resource_type_str = interaction.resource_type.get_resource_type()

    if iscoroutinefunction(interaction.handler):

        async def update_async(
            request: Request,
            response: Response,
            resource: ResourceType,
            id_: Id = Path(
                alias="id",
                description=Resource.schema()["properties"]["id"]["title"],
            ),
            _format: str = FORMAT_QP,
            _pretty: str = PRETTY_QP,
        ) -> Union[ResourceType, Response]:
            """Function for update interaction."""
            if resource and resource.id and id_ != resource.id:
                raise FHIRBadRequestError(
                    code="invalid",
                    details_text="Logical Id in URL must match logical Id in resource",
                )

            handler = cast(UpdateInteractionHandler[ResourceType], interaction.handler)
            result = await handler(InteractionContext(request, response), id_, resource)  # type: ignore[call-arg]
            _, result_resource = _result_to_id_resource_tuple(result)

            response.headers["Location"] = f"/{resource_type_str}/{id_}"

            return format_response(
                resource=result_resource,
                response=response,
                format_parameters=FormatParameters.from_request(request),
            )

        update_async.__annotations__.update(
            {
                "resource": interaction.resource_type,
            }
        )

        return update_async
    else:

        def update(
            request: Request,
            response: Response,
            resource: ResourceType,
            id_: Id = Path(
                alias="id",
                description=Resource.schema()["properties"]["id"]["title"],
            ),
            _format: str = FORMAT_QP,
            _pretty: str = PRETTY_QP,
        ) -> Union[ResourceType, Response]:
            """Function for update interaction."""
            if resource and resource.id and id_ != resource.id:
                raise FHIRBadRequestError(
                    code="invalid",
                    details_text="Logical Id in URL must match logical Id in resource",
                )

            handler = cast(UpdateInteractionHandler[ResourceType], interaction.handler)
            result = handler(InteractionContext(request, response), id_, resource)  # type: ignore[call-arg]
            _, result_resource = _result_to_id_resource_tuple(result)

            response.headers["Location"] = f"/{resource_type_str}/{id_}"

            return format_response(
                resource=result_resource,
                response=response,
                format_parameters=FormatParameters.from_request(request),
            )

        update.__annotations__.update(
            {
                "resource": interaction.resource_type,
            }
        )

        return update


def make_patch_function(
    interaction: TypeInteraction[ResourceType],
) -> Callable[
    [Request, Response, JSONPatch, Id, str, str],
    Union[
        Coroutine[None, None, Union[ResourceType, Response]],
        ResourceType,
        Response,
    ],
]:
    """Make a function suitable for creation of a FHIR patch API route."""
    resource_type_str = interaction.resource_type.get_resource_type()

    if iscoroutinefunction(interaction.handler):

        async def patch_async(
            request: Request,
            response: Response,
            json_patch: JSONPatch,
            id_: Id = Path(
                alias="id",
                description=Resource.schema()["properties"]["id"]["title"],
            ),
            _format: str = FORMAT_QP,
            _pretty: str = PRETTY_QP,
        ) -> Union[ResourceType, Response]:
            """Function for patch interaction."""
            handler = cast(PatchInteractionHandler[ResourceType], interaction.handler)
            result = await handler(InteractionContext(request, response), id_, json_patch)  # type: ignore[call-arg]
            _, result_resource = _result_to_id_resource_tuple(result)

            response.headers["Location"] = f"/{resource_type_str}/{id_}"

            return format_response(
                resource=result_resource,
                response=response,
                format_parameters=FormatParameters.from_request(request),
            )

        return patch_async
    else:

        def patch(
            request: Request,
            response: Response,
            json_patch: JSONPatch,
            id_: Id = Path(
                alias="id",
                description=Resource.schema()["properties"]["id"]["title"],
            ),
            _format: str = FORMAT_QP,
            _pretty: str = PRETTY_QP,
        ) -> Union[ResourceType, Response]:
            """Function for patch interaction."""
            handler = cast(PatchInteractionHandler[ResourceType], interaction.handler)
            result = handler(InteractionContext(request, response), id_, json_patch)  # type: ignore[call-arg]
            _, result_resource = _result_to_id_resource_tuple(result)

            response.headers["Location"] = f"/{resource_type_str}/{id_}"

            return format_response(
                resource=result_resource,
                response=response,
                format_parameters=FormatParameters.from_request(request),
            )

        return patch


def make_delete_function(
    interaction: TypeInteraction[ResourceType],
) -> Callable[
    [Request, Response, Id, str, str],
    Union[Coroutine[None, None, None], None],
]:
    """Make a function suitable for creation of a FHIR delete API route."""
    resource_type_str = interaction.resource_type.get_resource_type()

    if iscoroutinefunction(interaction.handler):

        async def delete_async(
            request: Request,
            response: Response,
            id_: Id = Path(
                alias="id",
                description=Resource.schema()["properties"]["id"]["title"],
            ),
            _format: str = FORMAT_QP,
            _pretty: str = PRETTY_QP,
        ) -> None:
            """Function for delete interaction."""
            handler = cast(DeleteInteractionHandler, interaction.handler)
            await handler(InteractionContext(request, response), id_)  # type: ignore[call-arg,misc]

            return None

        return delete_async
    else:

        def delete(
            request: Request,
            response: Response,
            id_: Id = Path(
                alias="id",
                description=Resource.schema()["properties"]["id"]["title"],
            ),
            _format: str = FORMAT_QP,
            _pretty: str = PRETTY_QP,
        ) -> None:
            """Function for delete interaction."""
            handler = cast(DeleteInteractionHandler, interaction.handler)
            handler(InteractionContext(request, response), id_)  # type: ignore[call-arg]

            return None

        return delete


def make_create_function(
    interaction: TypeInteraction[ResourceType],
) -> Callable[
    [Request, Response, ResourceType, str, str],
    Union[
        Coroutine[None, None, Union[ResourceType, Response]],
        ResourceType,
        Response,
    ],
]:
    """Make a function suitable for creation of a FHIR create API route."""
    resource_type_str = interaction.resource_type.get_resource_type()

    if iscoroutinefunction(interaction.handler):

        async def create_async(
            request: Request,
            response: Response,
            resource: ResourceType,
            _format: str = FORMAT_QP,
            _pretty: str = PRETTY_QP,
        ) -> Union[ResourceType, Response]:
            """
            Function for create interaction.

            Calls the handler, and sets the Location header based on the Id of the created resource.
            """
            handler = cast(CreateInteractionHandler[ResourceType], interaction.handler)
            result = await handler(InteractionContext(request, response), resource)  # type: ignore[call-arg]
            id_, result_resource = _result_to_id_resource_tuple(result)

            response.headers["Location"] = f"/{resource_type_str}/{id_}"

            return format_response(
                resource=result_resource,
                response=response,
                format_parameters=FormatParameters.from_request(request),
            )

        create_async.__annotations__.update(
            {
                "resource": interaction.resource_type,
            }
        )

        return create_async
    else:

        def create(
            request: Request,
            response: Response,
            resource: ResourceType,
            _format: str = FORMAT_QP,
            _pretty: str = PRETTY_QP,
        ) -> Union[ResourceType, Response]:
            """
            Function for create interaction.

            Calls the handler, and sets the Location header based on the Id of the created resource.
            """
            handler = cast(CreateInteractionHandler[ResourceType], interaction.handler)
            result = handler(InteractionContext(request, response), resource)  # type: ignore[call-arg]
            id_, result_resource = _result_to_id_resource_tuple(result)

            response.headers["Location"] = f"/{resource_type_str}/{id_}"

            return format_response(
                resource=result_resource,
                response=response,
                format_parameters=FormatParameters.from_request(request),
            )

        create.__annotations__.update(
            {
                "resource": interaction.resource_type,
            }
        )

        return create


# TODO: If possible, map FHIR primitives to correct type annotations for better validation
def make_search_type_function(
    interaction: TypeInteraction[ResourceType],
    search_parameter_metadata: Dict[str, Dict[str, str]],
    post: bool,
) -> Callable[
    [Request, Response, str, str],
    Union[Coroutine[None, None, Union[Bundle, Response]], Bundle, Response],
]:
    """
    Make a function suitable for creation of a FHIR search-type API route.

    Creation of a search-type function is more complex than creation of a create, read, or update
    function due to the variability of search parameters, support for custom search parameters, and
    due to the need to support GET and POST.

    Search parameter descriptions are pulled from the FHIR specification.

    After the function is created, the function signature is changed to account for what search
    parameters are supported by the developer-defined handler.
    """
    if post:
        format_annotation = Form(None, description=_FORMAT_PARAMETER_DESCRIPTION)
        pretty_annotation = Form(None, description=_PRETTY_PARAMETER_DESCRIPTION)
    else:
        format_annotation = FORMAT_QP
        pretty_annotation = PRETTY_QP

    search_parameters: Tuple[Parameter, ...] = tuple(
        _make_search_parameter(
            name=search_parameter.name,
            description=search_parameter_metadata[
                var_name_to_qp_name(search_parameter.name)
            ]["description"],
            post=post,
            multiple=search_parameter.multiple,
        )
        for search_parameter in supported_search_parameters(interaction.handler)
    )

    if iscoroutinefunction(interaction.handler):

        async def search_type_async(
            request: Request,
            response: Response,
            *,
            _format: str = format_annotation,
            _pretty: str = pretty_annotation,
            **kwargs: str,
        ) -> Union[Bundle, Response]:
            """Function for search-type interaction."""
            handler = cast(SearchTypeInteractionHandler, interaction.handler)
            bundle = await handler(InteractionContext(request, response), **kwargs)  # type: ignore[call-arg]

            return format_response(
                resource=bundle,
                response=response,
                format_parameters=FormatParameters.from_request(request),
            )

        return _set_search_type_function_signature(
            search_type_async, search_parameters, search_parameter_metadata
        )

    else:

        def search_type(
            request: Request,
            response: Response,
            *,
            _format: str = format_annotation,
            _pretty: str = pretty_annotation,
            **kwargs: str,
        ) -> Union[Bundle, Response]:
            """Function for search-type interaction."""
            handler = cast(SearchTypeInteractionHandler, interaction.handler)
            bundle = handler(InteractionContext(request, response), **kwargs)  # type: ignore[call-arg]

            return format_response(
                resource=bundle,
                response=response,
                format_parameters=FormatParameters.from_request(request),
            )

        return _set_search_type_function_signature(
            search_type, search_parameters, search_parameter_metadata
        )


def _result_to_id_resource_tuple(
    result: Union[Id, ResourceType],
) -> Tuple[Union[Id, None], Union[ResourceType, None]]:
    """
    Given an Id or a Resource, return an Id and a Resource.

    If a Resource is provided, both an Id and a Resource can be returned, however if an Id is
    provided, then only the Id can be returned.
    """
    if isinstance(result, Resource):
        return result.id, result
    else:
        return result, None


def _make_search_parameter(
    name: str, description: str, post: bool, multiple: bool
) -> Parameter:
    """
    Make a search parameter for the purpose of creating a function signature.

    Set the defaults to Form for POST endpoints, and Query for non-POST endpoints.
    """
    assert _is_valid_parameter_name(
        name
    ), f"{name} is not a valid search parameter name"

    return Parameter(
        name=name,
        kind=Parameter.KEYWORD_ONLY,
        default=(
            Form(None, alias=var_name_to_qp_name(name), description=description)
            if post
            else Query(None, alias=var_name_to_qp_name(name), description=description)
        ),
        annotation=List[str] if multiple else str,
    )


def _is_valid_parameter_name(name: str) -> bool:
    """
    Return True or False depending on whether the parameter name is valid.

    Names that are Python keywords are forbidden, in addition to other names that have additional
    meaning in Python or this package.
    """
    return not keyword.iskeyword(name) and name not in {
        "context",
        "format",
        "resource",
        "type",
    }


def _set_search_type_function_signature(
    search_type_function: Callable[
        ..., Union[Coroutine[None, None, Union[Bundle, Response]], Bundle, Response]
    ],
    search_parameters: Tuple[Parameter, ...],
    search_parameter_metadata: Dict[str, Dict[str, str]],
) -> Callable[
    [Request, Response, str, str],
    Union[Coroutine[None, None, Union[Bundle, Response]], Bundle, Response],
]:
    """
    Set the function signature of the search-type function so that it includes the search parameters
    that the handler supports.
    """
    sig = signature(search_type_function)
    parameters: Tuple[Parameter, ...] = tuple(sig.parameters.values())[:-1]

    sorted_search_parameters: List[Parameter] = sorted(
        parameters + search_parameters,
        key=lambda p: search_parameter_sort_key(
            p.name, search_parameter_metadata, p.annotation
        ),
    )

    sig = sig.replace(parameters=sorted_search_parameters)
    setattr(search_type_function, "__signature__", sig)

    return search_type_function



"""Test the capability statement"""

from typing import Any, Mapping, MutableMapping, Sequence, cast

import pytest

from .. import status
from ..fhir_specification import FHIR_SEQUENCE, FHIR_VERSION
from ..fhirstarter import FHIRStarter
from ..resources import CapabilityStatement
from ..testclient import TestClient
from .utils import assert_expected_response


@pytest.mark.parametrize(
    argnames="client,resource",
    argvalues=[
        (
            ["read", "update", "patch", "delete", "create", "search-type"],
            [
                {
                    "type": "Patient",
                    "interaction": [
                        {"code": "read"},
                        {"code": "update"},
                        {"code": "patch"},
                        {"code": "delete"},
                        {"code": "create"},
                        {"code": "search-type"},
                    ],
                    "searchParam": [
                        {
                            "name": "family",
                            "definition": "http://hl7.org/fhir/SearchParameter/individual-family",
                            "type": "string",
                            "documentation": "A portion of the family name of the patient",
                        },
                        {
                            "name": "general-practitioner",
                            "definition": "http://hl7.org/fhir/SearchParameter/"
                            "Patient-general-practitioner",
                            "type": "reference",
                            "documentation": "Patient's nominated general practitioner, not the "
                            "organization that manages the record",
                        },
                        {
                            "name": "nickname",
                            "definition": "https://hostname/nickname",
                            "type": "string",
                            "documentation": "Nickname",
                        },
                        {
                            "name": "_lastUpdated",
                            "definition": "http://hl7.org/fhir/SearchParameter/"
                            "Resource-lastUpdated",
                            "type": "date",
                            "documentation": "When the resource version last changed",
                        },
                    ],
                }
            ],
        ),
        (
            ["read", "create"],
            [
                {
                    "type": "Patient",
                    "interaction": [{"code": "read"}, {"code": "create"}],
                },
            ],
        ),
    ],
    ids=["all", "read and create"],
    indirect=["client"],
)
def test_capability_statement(
    client: TestClient, resource: Sequence[Mapping[str, Any]]
) -> None:
    """
    Test the capability statement.

    Two scenarios are parameterized: a server with all interaction types supported, and a server
    with only create and read supported.
    """
    app = cast(FHIRStarter, client.app)

    response = client.get("/metadata")

    assert_expected_response(
        response,
        status.HTTP_200_OK,
        content=_fhir_sequence_adjust(
            {
                "resourceType": "CapabilityStatement",
                "status": "active",
                "date": app._created,
                "kind": "instance",
                "fhirVersion": FHIR_VERSION,
                "acceptUnknown": "no",
                "format": ["json"],
                "rest": [
                    {
                        "mode": "server",
                        "resource": resource,
                    }
                ],
            }
        ),
    )


def test_capability_statement_pretty(client_create_and_read: TestClient) -> None:
    """Test the capability statement with a pretty response."""
    response = client_create_and_read.get("/metadata?_pretty=true")

    assert_expected_response(
        response,
        status.HTTP_200_OK,
        content=CapabilityStatement(**response.json()).json(
            indent=2, separators=(", ", ": ")
        ),
    )


@pytest.mark.parametrize(
    argnames="pretty",
    argvalues=["false", "true"],
    ids=["minified", "pretty"],
)
def test_capability_statement_xml(
    client_create_and_read: TestClient, pretty: str
) -> None:
    """Test the capability statement with an XML response."""
    response = client_create_and_read.get(f"/metadata?_format=xml&_pretty={pretty}")

    assert_expected_response(
        response,
        status.HTTP_200_OK,
        content_type="application/fhir+xml",
        content=CapabilityStatement.parse_raw(
            response.content, content_type="text/xml"
        ).xml(pretty_print=(pretty == "true")),
    )


def test_set_capability_statement_modifier(client_create_and_read: TestClient) -> None:
    """Test the set_capability_statement_modifier method."""
    app = cast(FHIRStarter, client_create_and_read.app)

    def modify_capability_statement(
        capability_statement: MutableMapping[str, Any], *_: Any
    ) -> MutableMapping[str, Any]:
        capability_statement["publisher"] = "Publisher"
        return capability_statement

    app.set_capability_statement_modifier(modify_capability_statement)

    response = client_create_and_read.get("/metadata")

    assert_expected_response(
        response,
        status.HTTP_200_OK,
        content=_fhir_sequence_adjust(
            {
                "resourceType": "CapabilityStatement",
                "status": "active",
                "date": app._created,
                "publisher": "Publisher",
                "kind": "instance",
                "fhirVersion": FHIR_VERSION,
                "acceptUnknown": "no",
                "format": ["json"],
                "rest": [
                    {
                        "mode": "server",
                        "resource": [
                            {
                                "type": "Patient",
                                "interaction": [{"code": "read"}, {"code": "create"}],
                            }
                        ],
                    }
                ],
            }
        ),
    )


def _fhir_sequence_adjust(
    capability_statement: MutableMapping[str, Any]
) -> MutableMapping[str, Any]:
    """
    Adjust a capability statement for the purposes of comparison.
    Example: For R4, R4B, and R5, the "acceptUnknown" value is no longer present.
    """
    if FHIR_SEQUENCE != "STU3":
        del capability_statement["acceptUnknown"]

    return capability_statement



"""
An example FHIR server implementation using FHIRStarter, with examples showing how to create FHIR
interactions (i.e. endpoints) that perform create, read, search-type, and update operations.
"""

import contextlib
import json
from pathlib import Path
from typing import Any, Dict, List, MutableMapping, Union
from uuid import uuid4

import jsonpatch
import uvicorn
from fhir.resources.bundle import Bundle
from fhir.resources.fhirtypes import Id
from fhir.resources.patient import Patient
from fhir.resources.practitioner import Practitioner
from starlette.responses import RedirectResponse

import fhirstarter
from fhirstarter import (
    FHIRProvider,
    FHIRStarter,
    InteractionContext,
    JSONPatch,
    Request,
    Response,
    convert_json_patch,
    examples,
)
from fhirstarter.exceptions import (
    FHIRResourceNotFoundError,
    FHIRUnprocessableEntityError,
)

# Create the app with the provided config file
app = FHIRStarter(
    title="FHIRStarter Example Implementation",
    config_file=Path(examples.__file__).parent / "config.toml",
)

# Create a "database"
DATABASE: Dict[str, Dict[str, str]] = {"Patient": {}, "Practitioner": {}}

# Create a provider
provider = FHIRProvider()

# To register FHIR interactions with  a provider, the pieces of information the developer has to
# provide are:
# * FHIR interaction type (this affects what the endpoint will look like)
# * FHIR resource type (this affects validation of inputs and outputs, and what search parameters
#   are valid)
# * the handler (expected signature and annotations need to match the defined protocols, because
#   these values affect route creation in FastAPI)


# Register the patient read FHIR interaction with the provider
@provider.read(Patient)
async def patient_read(context: InteractionContext, id_: Id) -> Patient:
    patient = DATABASE["Patient"].get(id_)
    if not patient:
        raise FHIRResourceNotFoundError

    return Patient(**json.loads(patient))


# Register the patient update FHIR interaction with the provider
@provider.update(Patient)
async def patient_update(context: InteractionContext, id_: Id, resource: Patient) -> Id:
    if id_ not in DATABASE["Patient"]:
        raise FHIRResourceNotFoundError

    DATABASE["Patient"][id_] = resource.json(separators=(",", ":"))

    return Id(id_)


# Register the patient patch FHIR interaction with the provider
@provider.patch(Patient)
async def patient_patch(
    context: InteractionContext, id_: Id, json_patch: JSONPatch
) -> Id:
    patient = json.loads(DATABASE["Patient"].get(id_, "{}"))
    if not patient:
        raise FHIRResourceNotFoundError

    # Convert the JSONPatch object to a list of dicts using the helper function, and use the
    # jsonpatch package to apply the patch to the patient resource
    patch = convert_json_patch(json_patch)
    jsonpatch.apply_patch(patient, jsonpatch.JsonPatch(patch), in_place=True)

    # Validate the change
    try:
        Patient.validate(patient)
    except Exception as exception:
        raise FHIRUnprocessableEntityError(
            code="invalid", details_text="Validation of patched resource failed"
        ) from exception

    DATABASE["Patient"][id_] = json.dumps(patient, separators=(",", ":"))

    return Id(id_)


@provider.delete(Patient)
async def patient_delete(context: InteractionContext, id_: Id) -> None:
    with contextlib.suppress(KeyError):
        del DATABASE["Patient"][id_]

    return None


# Register the patient create FHIR interaction with the provider
@provider.create(Patient)
async def patient_create(context: InteractionContext, resource: Patient) -> Id:
    id_ = str(uuid4())

    resource.id = id_
    DATABASE["Patient"][id_] = resource.json(separators=(",", ":"))

    return Id(id_)


# Register the patient search-type FHIR interaction with the provider
@provider.search_type(Patient)
async def patient_search_type(
    context: InteractionContext,
    birthdate: Union[List[str], None],
    general_practitioner: Union[str, None],
    family: Union[str, None],
    nickname: Union[str, None],
    _last_updated: Union[str, None],
) -> Bundle:
    patients = []
    for patient_serialized in DATABASE["Patient"].values():
        patient = json.loads(patient_serialized)
        for name in patient.get("name", {}):
            if name.get("family") == family:
                patients.append(patient)

    bundle = Bundle(
        **{
            "type": "searchset",
            "total": len(patients),
            "entry": [{"resource": patient} for patient in patients],
        }
    )

    return bundle


# Optional: Provide a custom example for the automatic documentation by defining a subclass of the
# FHIR Practitioner Pydantic model. If a custom model is not provided, examples from the FHIR
# specification are used.
class PractitionerCustom(Practitioner):
    class Config:
        schema_extra = {
            "example": {
                "resourceType": "Practitioner",
                "id": "example",
                "name": [{"family": "Careful", "given": ["Adam"], "prefix": ["Dr"]}],
            }
        }


# Register the practitioner read FHIR interaction with the provider using the custom subclass
@provider.read(PractitionerCustom)
async def practitioner_read(context: InteractionContext, id_: Id) -> PractitionerCustom:
    practitioner = DATABASE["Practitioner"].get(id_)
    if not practitioner:
        raise FHIRResourceNotFoundError

    return PractitionerCustom(**json.loads(practitioner))


# Add the provider to the app. This will automatically generate the API routes for the interactions
# provided by the providers (e.g. create, read, search-type, and update).
app.add_providers(provider)


# Customize the capability statement
def amend_capability_statement(
    capability_statement: MutableMapping[str, Any], request: Request, response: Response
) -> MutableMapping[str, Any]:
    capability_statement["publisher"] = "Canvas Medical"
    return capability_statement


app.set_capability_statement_modifier(amend_capability_statement)


# Redirect the main page to the API docs
@app.get("/", include_in_schema=False)
async def index() -> RedirectResponse:
    return RedirectResponse("/docs")


if __name__ == "__main__":
    # Start the server
    fhirstarter_dir = Path(fhirstarter.__file__).parent
    uvicorn.run(
        "example:app",
        use_colors=True,
        reload=True,
        reload_dirs=str(fhirstarter_dir.parent),
    )



"""
Resources module that finesses imports of resources used by the framework so that they can be
referenced by the same names across the testing framework regardless of the FHIR version being used.

The rationale behind this is the same as the one for the resources.py file that lives at the package
root.
"""

import os

FHIR_SEQUENCE = os.getenv("FHIR_SEQUENCE", "R5")

if FHIR_SEQUENCE in ("R4", "R5"):
    from fhir.resources.appointment import Appointment
    from fhir.resources.bundle import Bundle
    from fhir.resources.fhirtypes import Id
    from fhir.resources.humanname import HumanName
    from fhir.resources.patient import Patient
    from fhir.resources.practitioner import Practitioner
elif FHIR_SEQUENCE == "STU3":
    from fhir.resources.STU3.appointment import Appointment
    from fhir.resources.STU3.bundle import Bundle
    from fhir.resources.STU3.fhirtypes import Id
    from fhir.resources.STU3.humanname import HumanName
    from fhir.resources.STU3.patient import Patient
    from fhir.resources.STU3.practitioner import Practitioner
elif FHIR_SEQUENCE == "R4B":
    from fhir.resources.R4B.appointment import Appointment
    from fhir.resources.R4B.bundle import Bundle
    from fhir.resources.R4B.fhirtypes import Id
    from fhir.resources.R4B.humanname import HumanName
    from fhir.resources.R4B.patient import Patient
    from fhir.resources.R4B.practitioner import Practitioner

__all__ = ["Patient", "HumanName", "Appointment", "Bundle", "Id", "Practitioner"]



"""Utility functions for creation of routes and responses."""

from dataclasses import dataclass
from typing import Any, Callable, ClassVar, Dict, Literal, Union

from fastapi import Request
from fastapi.responses import JSONResponse, Response

from . import status
from .fhir_specification.utils import is_resource_type
from .interactions import ResourceType, SearchTypeInteraction, TypeInteraction
from .resources import Bundle, OperationOutcome, Resource


@dataclass
class InteractionInfo:
    resource_type: Union[str, None]
    interaction_type: Union[
        Literal[
            "read", "update", "patch", "delete", "create", "search-type", "capabilities"
        ],
        None,
    ]
    resource_id: Union[str, None]


def parse_fhir_request(request: Request) -> InteractionInfo:
    """
    Parse a FHIR request into its component parts, and determine an interaction type.

    Note: This function is currently oriented around specific use cases for this framework.
    Specifically, it will correctly categorize read, update, patch, create, search-type, and
    capabilities interactions. Further enhancement is needed to support more use cases.
    """
    no_info = InteractionInfo(  # type: ignore[call-arg]
        resource_type=None, interaction_type=None, resource_id=None
    )

    split_path = request.url.path.split("/")
    if not split_path:
        return no_info

    resource_id = None

    if request.method == "GET":
        if split_path[-1] == "metadata":
            return InteractionInfo(  # type: ignore[call-arg]
                resource_type=None, interaction_type="capabilities", resource_id=None
            )
        elif is_resource_type(split_path[-1]):
            resource_type = split_path[-1]
            interaction_type = "search-type"
        elif len(split_path) >= 3:
            resource_type, resource_id = split_path[-2:]
            interaction_type = "read"
        else:
            return no_info
    elif request.method == "POST":
        if is_resource_type(split_path[-1]):
            resource_type = split_path[-1]
            interaction_type = "create"
        elif split_path[-1] == "_search":
            resource_type = split_path[-2]
            interaction_type = "search-type"
        else:
            return no_info
    elif request.method == "PUT":
        if len(split_path) >= 3:
            resource_type, resource_id = split_path[-2:]
            interaction_type = "update"
        else:
            return no_info
    elif request.method == "PATCH":
        if len(split_path) >= 3:
            resource_type, resource_id = split_path[-2:]
            interaction_type = "patch"
        else:
            return no_info
    elif request.method == "DELETE":
        if len(split_path) >= 3:
            resource_type, resource_id = split_path[-2:]
            interaction_type = "delete"
        else:
            return no_info
    else:
        return no_info

    if not is_resource_type(resource_type):
        return no_info

    return InteractionInfo(  # type: ignore[call-arg]
        resource_type=resource_type,
        interaction_type=interaction_type,
        resource_id=resource_id,
    )


def make_operation_outcome(
    severity: str, code: str, details_text: str
) -> OperationOutcome:
    """Create a simple OperationOutcome given a severity, code, and details."""
    return OperationOutcome(
        **{
            "issue": [
                {
                    "severity": severity,
                    "code": code,
                    "details": {"text": details_text},
                }
            ]
        }
    )


@dataclass
class FormatParameters:
    format: str = "application/fhir+json"
    pretty: bool = False

    _CONTENT_TYPES: ClassVar = {
        "json": "application/fhir+json",
        "application/json": "application/fhir+json",
        "application/fhir+json": "application/fhir+json",
        "xml": "application/fhir+xml",
        "text/xml": "application/fhir+xml",
        "application/xml": "application/fhir+xml",
        "application/fhir+xml": "application/fhir+xml",
    }

    @classmethod
    def from_request(
        cls, request: Request, raise_exception: bool = True
    ) -> "FormatParameters":
        """
        Parse the _format and _pretty query parameters.

        The value for format is first obtained from the Accept header, and if not specified there is
        obtained from the _format query parameter.
        """
        format_ = cls.format_from_accept_header(request)

        try:
            if not format_:
                format_ = cls._CONTENT_TYPES[
                    request.query_params.get("_format", "json")
                ]
        except KeyError:
            if raise_exception:
                from .exceptions import FHIRGeneralError

                raise FHIRGeneralError(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    severity="error",
                    code="structure",
                    details_text="Invalid response format specified for '_format' parameter",
                )
            else:
                format_ = "application/fhir+json"

        return cls(  # type: ignore[call-arg]
            format=format_,
            pretty=request.query_params.get("_pretty", "false") == "true",
        )

    @classmethod
    def format_from_accept_header(cls, request: Request) -> Union[str, None]:
        if request.method == "POST":
            for content_type in request.headers.getlist("Accept"):
                if content_type_normalized := cls._CONTENT_TYPES.get(content_type):
                    return content_type_normalized

        return None


def format_response(
    resource: Union[Resource, None],
    response: Union[Response, None] = None,
    status_code: Union[int, None] = None,
    format_parameters: FormatParameters = FormatParameters(),
) -> Union[Resource, Response]:
    """
    Return a response with the proper formatting applied.

    This function provides a response in JSON or XML format that has been prettified if requested.

    There are six scenarios that are handled:
    1. Null resource (when there is no body -- no handling required)
    1. Pretty JSON
    2. Minified JSON with a status code (mainly for errors)
    3. Minified JSON with no specified status code (usually the default)
    4. Pretty XML
    5. Minified XML
    """
    if not resource:
        assert (
            response is not None
        ), "Response object must be provided for a null resource"
        response.headers["Content-Type"] = format_parameters.format
        return resource

    if format_parameters.format == "application/fhir+json":
        if format_parameters.pretty:
            return Response(
                content=resource.json(
                    ensure_ascii=False,
                    allow_nan=False,
                    indent=2,
                    separators=(", ", ": "),
                ),
                status_code=status_code or status.HTTP_200_OK,
                media_type=format_parameters.format,
            )
        else:
            if status_code:
                return JSONResponse(
                    content=resource.dict(),
                    status_code=status_code,
                    media_type=format_parameters.format,
                )
            else:
                assert (
                    response is not None
                ), "Response object or status code must be provided for non-pretty JSON responses"
                response.headers["Content-Type"] = format_parameters.format
                return resource
    else:
        return Response(
            content=resource.xml(pretty_print=format_parameters.pretty),
            status_code=status_code or status.HTTP_200_OK,
            media_type=format_parameters.format,
        )


def read_route_args(interaction: TypeInteraction[ResourceType]) -> Dict[str, Any]:
    """Provide arguments for creation of a FHIR read API route."""
    resource_type_str = interaction.resource_type.get_resource_type()

    return {
        "path": f"/{resource_type_str}/{{id}}",
        "response_model": interaction.resource_type,
        "status_code": status.HTTP_200_OK,
        "tags": [f"Type:{resource_type_str}"],
        "summary": f"{resource_type_str} {interaction.label()}",
        "description": f"The {resource_type_str} read interaction accesses "
        f"the current contents of a {resource_type_str} resource.",
        "responses": _responses(
            interaction,
            _ok,
            _unauthorized,
            _forbidden,
            _not_found,
            _internal_server_error,
        ),
        "operation_id": f"fhirstarter|instance|read|get|{resource_type_str}|{interaction.resource_type.__module__}|{interaction.resource_type.__name__}",
        "response_model_exclude_none": True,
        **interaction.route_options,
    }


def update_route_args(interaction: TypeInteraction[ResourceType]) -> Dict[str, Any]:
    """Provide arguments for creation of a FHIR update API route."""
    resource_type_str = interaction.resource_type.get_resource_type()

    return {
        "path": f"/{resource_type_str}/{{id}}",
        "response_model": Union[interaction.resource_type, None],
        "status_code": status.HTTP_200_OK,
        "tags": [f"Type:{resource_type_str}"],
        "summary": f"{resource_type_str} {interaction.label()}",
        "description": f"The {resource_type_str} update interaction creates a new current version "
        f"for an existing {resource_type_str} resource.",
        "responses": _responses(
            interaction,
            _ok,
            _bad_request,
            _unauthorized,
            _forbidden,
            _not_found,
            _unprocessable_entity,
            _internal_server_error,
        ),
        "operation_id": f"fhirstarter|instance|update|put|{resource_type_str}|{interaction.resource_type.__module__}|{interaction.resource_type.__name__}",
        "response_model_exclude_none": True,
        **interaction.route_options,
    }


def patch_route_args(interaction: TypeInteraction[ResourceType]) -> Dict[str, Any]:
    """Provide arguments for creation of a FHIR patch API route."""
    resource_type_str = interaction.resource_type.get_resource_type()

    return {
        "path": f"/{resource_type_str}/{{id}}",
        "response_model": Union[interaction.resource_type, None],
        "status_code": status.HTTP_200_OK,
        "tags": [f"Type:{resource_type_str}"],
        "summary": f"{resource_type_str} {interaction.label()}",
        "description": f"The {resource_type_str} patch interaction creates a new current version "
        f"for an existing {resource_type_str} resource by applying operations described in a "
        "[JSON Patch document](https://datatracker.ietf.org/doc/html/rfc6902).",
        "responses": _responses(
            interaction,
            _ok,
            _bad_request,
            _unauthorized,
            _forbidden,
            _not_found,
            _unprocessable_entity,
            _internal_server_error,
        ),
        "operation_id": f"fhirstarter|instance|patch|patch|{resource_type_str}|{interaction.resource_type.__module__}|{interaction.resource_type.__name__}",
        "response_model_exclude_none": True,
        **interaction.route_options,
    }


def delete_route_args(interaction: TypeInteraction[ResourceType]) -> Dict[str, Any]:
    """Provide arguments for creation of a FHIR delete API route."""
    resource_type_str = interaction.resource_type.get_resource_type()

    return {
        "path": f"/{resource_type_str}/{{id}}",
        "response_model": None,
        "status_code": status.HTTP_204_NO_CONTENT,
        "tags": [f"Type:{resource_type_str}"],
        "summary": f"{resource_type_str} {interaction.label()}",
        "description": f"The {resource_type_str} delete interaction removes an existing "
        f"{resource_type_str} resource.",
        "responses": _responses(
            interaction,
            _no_content,
            _unauthorized,
            _forbidden,
            _internal_server_error,
        ),
        "operation_id": f"fhirstarter|instance|delete|delete|{resource_type_str}|{interaction.resource_type.__module__}|{interaction.resource_type.__name__}",
        "response_model_exclude_none": True,
        **interaction.route_options,
    }


def create_route_args(interaction: TypeInteraction[ResourceType]) -> Dict[str, Any]:
    """Provide arguments for creation of a FHIR create API route."""
    resource_type_str = interaction.resource_type.get_resource_type()

    return {
        "path": f"/{resource_type_str}",
        "response_model": Union[interaction.resource_type, None],
        "status_code": status.HTTP_201_CREATED,
        "tags": [f"Type:{resource_type_str}"],
        "summary": f"{resource_type_str} {interaction.label()}",
        "description": f"The {resource_type_str} create interaction creates a new "
        f"{resource_type_str} resource in a server-assigned location.",
        "responses": _responses(
            interaction,
            _created,
            _bad_request,
            _unauthorized,
            _forbidden,
            _unprocessable_entity,
            _internal_server_error,
        ),
        "operation_id": f"fhirstarter|type|create|post|{resource_type_str}|{interaction.resource_type.__module__}|{interaction.resource_type.__name__}",
        "response_model_exclude_none": True,
        **interaction.route_options,
    }


def search_type_route_args(
    interaction: TypeInteraction[ResourceType], post: bool
) -> Dict[str, Any]:
    """Provide arguments for creation of a FHIR search-type API route."""
    resource_type_str = interaction.resource_type.get_resource_type()

    return {
        "path": f"/{resource_type_str}{'/_search' if post else ''}",
        "response_model": Bundle,
        "status_code": status.HTTP_200_OK,
        "tags": [f"Type:{resource_type_str}"],
        "summary": f"{resource_type_str} {interaction.label()}",
        "description": f"The {resource_type_str} search-type interaction searches a set of "
        "resources based on some filter criteria.",
        "responses": _responses(
            interaction,
            _ok,
            _bad_request,
            _unauthorized,
            _forbidden,
            _internal_server_error,
        ),
        "operation_id": f"fhirstarter|type|search-type|{'post' if post else 'get'}|{resource_type_str}|{interaction.resource_type.__module__}|{interaction.resource_type.__name__}",
        "response_model_exclude_none": True,
        **interaction.route_options,
    }


_Responses = Dict[int, Dict[str, Any]]


def _responses(
    interaction: TypeInteraction[ResourceType],
    *responses: Callable[[TypeInteraction[ResourceType]], _Responses],
) -> _Responses:
    """Combine the responses documentation for a FHIR interaction into a single dictionary."""
    merged_responses: _Responses = {}
    for response in responses:
        merged_responses.update(response(interaction))
    return merged_responses


def _ok(interaction: TypeInteraction[ResourceType]) -> _Responses:
    """Return documentation for an HTTP 200 OK response."""
    return {
        status.HTTP_200_OK: {
            "model": (
                interaction.resource_type
                if not isinstance(interaction, SearchTypeInteraction)
                else Bundle
            ),
            "description": f"Successful {interaction.resource_type.get_resource_type()} "
            f"{interaction.label()}",
        }
    }


def _created(interaction: TypeInteraction[ResourceType]) -> _Responses:
    """Documentation for an HTTP 201 Created response."""
    return {
        status.HTTP_201_CREATED: {
            "model": interaction.resource_type,
            "description": f"Successful {interaction.resource_type.get_resource_type()} create",
        }
    }


def _no_content(interaction: TypeInteraction[ResourceType]) -> _Responses:
    """Return documentation for an HTTP 204 No Content response."""
    return {
        status.HTTP_204_NO_CONTENT: {
            "model": None,
            "description": f"Successful {interaction.resource_type.get_resource_type()} "
            f"{interaction.label()}",
        }
    }


def _bad_request(interaction: TypeInteraction[ResourceType]) -> _Responses:
    """Documentation for an HTTP 400 Bad Request response."""
    return {
        status.HTTP_400_BAD_REQUEST: {
            "model": OperationOutcome,
            "description": f"{interaction.resource_type.get_resource_type()} "
            f"{interaction.label()} request could not be parsed or "
            "failed basic FHIR validation rules.",
        }
    }


def _unauthorized(interaction: TypeInteraction[ResourceType]) -> _Responses:
    """Documentation for an HTTP 401 Unauthorized response."""
    return {
        status.HTTP_401_UNAUTHORIZED: {
            "model": OperationOutcome,
            "description": "Authentication is required for the "
            f"{interaction.resource_type.get_resource_type()} {interaction.label()} interaction "
            "that was attempted.",
        }
    }


def _forbidden(interaction: TypeInteraction[ResourceType]) -> _Responses:
    """Documentation for an HTTP 403 Forbidden response."""
    return {
        status.HTTP_403_FORBIDDEN: {
            "model": OperationOutcome,
            "description": "Authorization is required for the "
            f"{interaction.resource_type.get_resource_type()} {interaction.label()} interaction "
            "that was attempted.",
        }
    }


def _not_found(interaction: TypeInteraction[ResourceType]) -> _Responses:
    """Documentation for an HTTP 404 Not Found response."""
    return {
        status.HTTP_404_NOT_FOUND: {
            "model": OperationOutcome,
            "description": f"Unknown {interaction.resource_type.get_resource_type()} resource",
        }
    }


def _unprocessable_entity(interaction: TypeInteraction[ResourceType]) -> _Responses:
    """Documentation for an HTTP 422 Unprocessable Entity response."""
    return {
        status.HTTP_422_UNPROCESSABLE_ENTITY: {
            "model": OperationOutcome,
            "description": f"The proposed {interaction.resource_type.get_resource_type()} resource"
            " violated applicable "
            "FHIR profiles or server business rules.",
        }
    }


def _internal_server_error(_: TypeInteraction[ResourceType]) -> _Responses:
    """Documentation for an HTTP 500 Internal Server error response."""
    return {
        status.HTTP_500_INTERNAL_SERVER_ERROR: {
            "model": OperationOutcome,
            "description": f"The server has encountered a situation it does not know how to handle.",
        }
    }



"""OpenAPI schema modifications"""

from collections import defaultdict
from dataclasses import dataclass
from importlib import import_module
from typing import (
    Any,
    DefaultDict,
    Dict,
    Iterator,
    List,
    Mapping,
    MutableMapping,
    Tuple,
    Union,
    cast,
)

from .fhir_specification.utils import (
    create_bundle_example,
    is_resource_type,
    load_examples,
    make_operation_outcome_example,
)
from .fhirstarter import status

__all__ = ["adjust_schema"]


@dataclass
class _OperationId:
    interaction_type: str
    method: str
    module_name: Union[str, None]
    model_name: Union[str, None]


def _parse_operation_id(operation_id: str) -> _OperationId:
    """Return a parsed operation ID."""
    tokens: List[str] = operation_id.split("|")

    interaction_type = tokens[2]
    method = tokens[3]
    module_name = tokens[5] if len(tokens) > 5 else None
    model_name = tokens[6] if len(tokens) > 6 else None

    return _OperationId(interaction_type, method, module_name, model_name)  # type: ignore[call-arg]


def _operations(
    openapi_schema: MutableMapping[str, Any]
) -> Iterator[Tuple[_OperationId, Dict[str, Any]]]:
    """Yield operations in the OpenAPI schema that were created by FHIRStarter."""
    for path_name, path in openapi_schema["paths"].items():
        for operation in path.values():
            operation_id = operation.get("operationId", "")
            if operation_id.startswith("fhirstarter|"):
                yield _parse_operation_id(operation_id), operation


def _search_type_operations(
    openapi_schema: MutableMapping[str, Any]
) -> Iterator[Tuple[_OperationId, Dict[str, Any]]]:
    """Yield search-type operations in the OpenAPI schema that were created by FHIRStarter."""
    for operation_id, operation in _operations(openapi_schema):
        if operation_id.interaction_type == "search-type":
            yield operation_id, operation


def adjust_schema(openapi_schema: MutableMapping[str, Any]) -> None:
    """
    Adjust the OpenAPI schema to make it more FHIR-friendly.

    Remove some default schemas that are not needed nor used, and change all content types that
    are set to "application/json" to instead be "application/fhir+json". Make a few additional
    aesthetic changes to clean up the auto-generated documentation.

    Because it directly modifies the OpenAPI schema, it is vulnerable to breakage from updates
    to FastAPI. This is not a significant vulnerability because the core server functionality
    will still work (i.e. this is just documentation).
    """
    _inline_search_post_schemas(openapi_schema)

    # Add any missing schemas (from search-type interactions). This needs to run before the rest of
    # the tasks so that all the schemas are there for subsequent actions.
    _add_schemas(openapi_schema)

    examples = _get_examples(openapi_schema)
    for operation_id, operation in _operations(openapi_schema):
        _adjust_operation(operation_id, operation, examples)


def _inline_search_post_schemas(openapi_schema: MutableMapping[str, Any]) -> None:
    """
    Inline the schemas generated for search by POST. These schemas are only used in one place, so
    they don't need to exist in the schemas section.
    """
    for operation_id, operation in _search_type_operations(openapi_schema):
        # Copy and inline the schema, and remove it from the schemas section
        if (
            operation_id.interaction_type == "search-type"
            and operation_id.method == "post"
        ):
            request_body = operation["requestBody"]
            schema_name = request_body["content"]["application/x-www-form-urlencoded"][
                "schema"
            ]["$ref"].split("/")[-1]
            request_body["content"]["application/x-www-form-urlencoded"]["schema"] = (
                openapi_schema["components"]["schemas"].pop(schema_name)
            )


def _add_schemas(openapi_schema: MutableMapping[str, Any]) -> None:
    """
    Add missing schemas.

    If a server only supports search for a given resource, then the OpenAPI schema won't include the
    schema for the resources that are returned in the bundle by the search interaction.
    """
    for operation_id, operation in _search_type_operations(openapi_schema):
        if (
            operation_id.model_name
            and operation_id.model_name not in openapi_schema["components"]["schemas"]
        ):
            module = import_module(operation_id.module_name)
            model = getattr(module, operation_id.model_name)
            openapi_schema["components"]["schemas"][
                operation_id.model_name
            ] = model.schema()

    # Recreate the schema dictionary so that the schemas appear sorted
    openapi_schema["components"]["schemas"] = {
        schema_name: schema
        for schema_name, schema in sorted(
            openapi_schema["components"]["schemas"].items()
        )
    }


def _get_examples(
    openapi_schema: MutableMapping[str, Any]
) -> Dict[str, Dict[str, Any]]:
    """
    Gather examples for all scenarios: request and response bodies for interactions;
    resource-specific Bundle examples for search interactions; and OperationOutcome examples for
    errors.
    """
    examples: DefaultDict[str, Any] = defaultdict(dict)

    # Get all resource examples from the models and the FHIR specification
    for schema_name, schema in openapi_schema["components"]["schemas"].items():
        properties = schema.get("properties", {})
        if "resource_type" not in properties:
            continue

        # Bundle, OperationOutcome, and Parameters are handled differently
        resource_type = properties["resource_type"]["const"]
        if resource_type in {"Bundle", "OperationOutcome", "Parameters"}:
            continue

        # If there is a custom example on the model, use it. Otherwise, get examples from the FHIR
        # specification. Add the example(s) and a bundle example.
        if schema_example := schema.get("example"):
            examples[schema_name]["example"] = schema_example
            examples["Bundle"][schema_name] = create_bundle_example(schema_example)
        elif is_resource_type(resource_type):
            resource_examples = load_examples(resource_type)
            examples[schema_name]["examples"] = resource_examples
            examples["Bundle"][schema_name] = create_bundle_example(
                next(iter(resource_examples.values()))["value"]
            )
        else:
            resource_example = {"resourceType": resource_type}
            examples[schema_name]["example"] = resource_example
            examples["Bundle"][schema_name] = create_bundle_example(resource_example)

    # Make OperationOutcome examples
    for status_code, code, details_text in (
        (str(status.HTTP_400_BAD_REQUEST), "invalid", "Bad request"),
        (str(status.HTTP_401_UNAUTHORIZED), "unknown", "Authentication failed"),
        (str(status.HTTP_403_FORBIDDEN), "forbidden", "Authorization failed"),
        (str(status.HTTP_404_NOT_FOUND), "not-found", "Resource not found"),
        (
            str(status.HTTP_422_UNPROCESSABLE_ENTITY),
            "processing",
            "Unprocessable entity",
        ),
        (
            str(status.HTTP_500_INTERNAL_SERVER_ERROR),
            "exception",
            "Internal server error",
        ),
    ):
        examples["OperationOutcome"][status_code] = make_operation_outcome_example(
            severity="error", code=code, details_text=details_text
        )

    return examples


def _adjust_operation(
    operation_id: _OperationId,
    operation: MutableMapping[str, Any],
    examples: Mapping[str, Dict[str, Any]],
) -> None:
    """
    Make adjustments to an operation in the OpenAPI schema.

    * Change the application/json content type to application/fhir+json
    * Remove the default FastAPI response schema (HTTPValidationError)
    * Add request and response body examples.
    """
    # Get the examples
    if operation_id.interaction_type == "capabilities":
        resource_examples = examples["CapabilityStatement"]
    else:
        resource_examples = examples[cast(str, operation_id.model_name)]

    # For operations that take a request body (excluding patch), change the application/json content
    # type to application/fhir+json and add request body examples
    if operation_id.interaction_type != "patch":
        if content := operation.get("requestBody", {}).get("content"):
            if "application/json" in content:
                content["application/fhir+json"] = content.pop("application/json")
                content["application/fhir+json"].update(resource_examples)

    # For each possible response (i.e. status code), remove the default FastAPI response schema
    responses = operation["responses"]
    status_codes: Tuple[str, ...] = tuple(responses.keys())
    for status_code in status_codes:
        if (
            status_code != str(status.HTTP_204_NO_CONTENT)
            and responses[status_code]["content"]
            .get("application/json", {})
            .get("schema", {})
            .get("$ref")
            == "#/components/schemas/HTTPValidationError"
        ):
            responses.pop(status_code)

    # For each response, change all instances of application/json to application/fhir+json and add
    # response body examples
    for status_code, response in responses.items():
        if status_code == str(status.HTTP_204_NO_CONTENT):
            continue

        # Move the response for "application/json" to "application/fhir+json"
        schema = response["content"].pop("application/json", None)
        if schema:
            response["content"]["application/fhir+json"] = schema

        # Add examples for success responses
        if 200 <= int(status_code) <= 299:
            if operation_id.interaction_type != "search-type":
                response["content"]["application/fhir+json"].update(resource_examples)
            elif operation_id.model_name:
                response["content"]["application/fhir+json"]["example"] = examples[
                    "Bundle"
                ][operation_id.model_name]

        # Add specialized OperationOutcome responses if available for the status code
        if operation_outcome_example := examples["OperationOutcome"].get(status_code):
            response["content"]["application/fhir+json"][
                "example"
            ] = operation_outcome_example



"""Test FHIR utils"""

import pytest

from ..utils import InteractionInfo, parse_fhir_request
from .utils import generate_fhir_resource_id, make_request


@pytest.mark.parametrize(
    argnames="mount_path",
    argvalues=["", "/subapi"],
    ids=["without mount", "with mount"],
)
@pytest.mark.parametrize(
    argnames="_,request_method,path,expected_result",
    argvalues=(
        argvalues := [
            (
                "capabilitites",
                "GET",
                "/metadata",
                InteractionInfo(  # type: ignore[call-arg]
                    resource_type=None,
                    interaction_type="capabilities",
                    resource_id=None,
                ),
            ),
            (
                "read",
                "GET",
                f"/Patient/{(id_ := generate_fhir_resource_id())}",
                InteractionInfo(  # type: ignore[call-arg]
                    resource_type="Patient", interaction_type="read", resource_id=id_
                ),
            ),
            (
                "read unrecognized resource type",
                "GET",
                f"/FakeResource/{(id_ := generate_fhir_resource_id())}",
                InteractionInfo(  # type: ignore[call-arg]
                    resource_type=None, interaction_type=None, resource_id=None
                ),
            ),
            (
                "update",
                "PUT",
                f"/Patient/{(id_ := generate_fhir_resource_id())}",
                InteractionInfo(  # type: ignore[call-arg]
                    resource_type="Patient", interaction_type="update", resource_id=id_
                ),
            ),
            (
                "update unrecognized resource type",
                "PUT",
                f"/FakeResource/{(id_ := generate_fhir_resource_id())}",
                InteractionInfo(  # type: ignore[call-arg]
                    resource_type=None, interaction_type=None, resource_id=None
                ),
            ),
            (
                "patch",
                "PATCH",
                f"/Patient/{(id_ := generate_fhir_resource_id())}",
                InteractionInfo(  # type: ignore[call-arg]
                    resource_type="Patient", interaction_type="patch", resource_id=id_
                ),
            ),
            (
                "patch unrecognized resource type",
                "PATCH",
                f"/FakeResource/{(id_ := generate_fhir_resource_id())}",
                InteractionInfo(  # type: ignore[call-arg]
                    resource_type=None, interaction_type=None, resource_id=None
                ),
            ),
            (
                "delete",
                "DELETE",
                f"/Patient/{(id_ := generate_fhir_resource_id())}",
                InteractionInfo(  # type: ignore[call-arg]
                    resource_type="Patient", interaction_type="delete", resource_id=id_
                ),
            ),
            (
                "delete unrecognized resource type",
                "DELETE",
                f"/FakeResource/{(id_ := generate_fhir_resource_id())}",
                InteractionInfo(  # type: ignore[call-arg]
                    resource_type=None, interaction_type=None, resource_id=None
                ),
            ),
            (
                "create",
                "POST",
                "/Patient",
                InteractionInfo(  # type: ignore[call-arg]
                    resource_type="Patient", interaction_type="create", resource_id=None
                ),
            ),
            (
                "create unrecognized resource type",
                "POST",
                "/FakeResource",
                InteractionInfo(  # type: ignore[call-arg]
                    resource_type=None, interaction_type=None, resource_id=None
                ),
            ),
            (
                "search-type",
                "GET",
                "/Patient",
                InteractionInfo(  # type: ignore[call-arg]
                    resource_type="Patient",
                    interaction_type="search-type",
                    resource_id=None,
                ),
            ),
            (
                "search-type post",
                "POST",
                "/Patient/_search",
                InteractionInfo(  # type: ignore[call-arg]
                    resource_type="Patient",
                    interaction_type="search-type",
                    resource_id=None,
                ),
            ),
            (
                "search unrecognized resource type",
                "GET",
                f"/FakeResource",
                InteractionInfo(  # type: ignore[call-arg]
                    resource_type=None, interaction_type=None, resource_id=None
                ),
            ),
            (
                "search POST unrecognized resource type",
                "POST",
                "/FakeResource/_search",
                InteractionInfo(  # type: ignore[call-arg]
                    resource_type=None, interaction_type=None, resource_id=None
                ),
            ),
            (
                "unrecognized GET path",
                "GET",
                f"/Patient/{(id_ := generate_fhir_resource_id())}/extra",
                InteractionInfo(  # type: ignore[call-arg]
                    resource_type=None, interaction_type=None, resource_id=None
                ),
            ),
            (
                "unrecognized PUT path",
                "PUT",
                f"/FakeResource/{(id_ := generate_fhir_resource_id())}/extra",
                InteractionInfo(  # type: ignore[call-arg]
                    resource_type=None, interaction_type=None, resource_id=None
                ),
            ),
            (
                "unrecognized POST path",
                "POST",
                "/FakeResource/extra",
                InteractionInfo(  # type: ignore[call-arg]
                    resource_type=None, interaction_type=None, resource_id=None
                ),
            ),
            (
                "unrecognized PATCH path",
                "PATCH",
                f"/FakeResource/{(id_ := generate_fhir_resource_id())}/extra",
                InteractionInfo(  # type: ignore[call-arg]
                    resource_type=None, interaction_type=None, resource_id=None
                ),
            ),
            (
                "unrecognized DELETE path",
                "DELETE",
                f"/FakeResource/{(id_ := generate_fhir_resource_id())}/extra",
                InteractionInfo(  # type: ignore[call-arg]
                    resource_type=None, interaction_type=None, resource_id=None
                ),
            ),
            (
                "unsupported HTTP method",
                "HEAD",
                f"/Patient/{(id_ := generate_fhir_resource_id())}/extra",
                InteractionInfo(  # type: ignore[call-arg]
                    resource_type=None, interaction_type=None, resource_id=None
                ),
            ),
        ]
    ),
    ids=[id_ for id_, *_ in argvalues],
)
def test_parse_fhir_request(
    _: str,
    mount_path: str,
    request_method: str,
    path: str,
    expected_result: InteractionInfo,
) -> None:
    assert (
        parse_fhir_request(make_request(request_method, f"{mount_path}{path}"))
        == expected_result
    )



"""Test FHIRStarter error handling"""

import json
from typing import Any, Callable, Coroutine, Mapping, Union, cast

import pytest
from fastapi import HTTPException

from ..exceptions import (
    FHIRBadRequestError,
    FHIRConflictError,
    FHIRForbiddenError,
    FHIRGoneError,
    FHIRMethodNotAllowedError,
    FHIRNotAcceptableError,
    FHIRPreconditionFailedError,
    FHIRUnauthorizedError,
    FHIRUnprocessableEntityError,
    FHIRUnsupportedMediaTypeError,
)
from ..fhirstarter import FHIRProvider, FHIRStarter, Request, Response, status
from ..testclient import TestClient
from ..utils import make_operation_outcome
from .config import app
from .resources import Patient
from .utils import assert_expected_response, generate_fhir_resource_id


@pytest.mark.parametrize(
    argnames="client,request_body,response_body",
    argvalues=[
        (
            ("read", "update", "patch", "delete", "create", "search-type"),
            " ",
            make_operation_outcome(
                severity="error",
                code="structure",
                details_text="body -> 1 — JSON decode error (type=json_invalid; error=Expecting value)",
            ),
        ),
        (
            ("read", "update", "patch", "delete", "create", "search-type"),
            {"extraField": []},
            make_operation_outcome(
                severity="error",
                code="structure",
                details_text="body -> extraField — extra fields not permitted "
                "(type=value_error.extra)",
            ),
        ),
        (
            ("read", "update", "patch", "delete", "create", "search-type"),
            {"communication": [{}]},
            make_operation_outcome(
                severity="error",
                code="required",
                details_text="body -> communication -> 0 -> language — field required "
                "(type=value_error.missing)",
            ),
        ),
        (
            ("read", "update", "patch", "delete", "create", "search-type"),
            {"id": ""},
            make_operation_outcome(
                severity="error",
                code="value",
                details_text="body -> id — ensure this value has at least 1 characters "
                "(type=value_error.any_str.min_length; limit_value=1)",
            ),
        ),
        (
            ("read", "update", "patch", "delete", "create", "search-type"),
            {"name": 0},
            make_operation_outcome(
                severity="error",
                code="value",
                details_text="body -> name — value is not a valid list (type=type_error.list)",
            ),
        ),
        (
            ("read", "update", "patch", "delete", "create", "search-type"),
            {"extraField": [], "communication": [{}]},
            {
                "resourceType": "OperationOutcome",
                "issue": [
                    {
                        "severity": "error",
                        "code": "required",
                        "details": {
                            "text": "body -> communication -> 0 -> language — field required "
                            "(type=value_error.missing)"
                        },
                    },
                    {
                        "severity": "error",
                        "code": "structure",
                        "details": {
                            "text": "body -> extraField — extra fields not permitted "
                            "(type=value_error.extra)"
                        },
                    },
                ],
            },
        ),
    ],
    ids=[
        "JSON decode error",
        "extra value",
        "missing value",
        "value error",
        "type error",
        "multiple errors",
    ],
    indirect=["client"],
)
def test_validation_error(
    client: TestClient,
    request_body: Union[Mapping[str, Any], str],
    response_body: Mapping[str, Any],
) -> None:
    """
    Test FHIR create interaction that produces 400 bad request error due to a validation failure.
    """
    if isinstance(request_body, Mapping):
        request_body = json.dumps(request_body)

    create_response = client.post("/Patient", content=request_body)

    assert_expected_response(
        create_response,
        status.HTTP_400_BAD_REQUEST,
        content=response_body,
    )


def _handler_exception_async(
    exception: HTTPException,
) -> Callable[..., Coroutine[None, None, Patient]]:
    """Return an async Patient read handler."""

    async def patient_read(*_: Any) -> Patient:
        raise exception

    return patient_read


def _handler_exception(exception: HTTPException) -> Callable[..., Patient]:
    """Return a Patient read handler."""

    def patient_read(*_: Any) -> Patient:
        raise exception

    return patient_read


@pytest.mark.parametrize(
    argnames="exception,status_code,issue",
    argvalues=[
        (
            HTTPException(status_code=status.HTTP_400_BAD_REQUEST),
            status.HTTP_400_BAD_REQUEST,
            {
                "severity": "error",
                "code": "processing",
                "details": {"text": "Bad Request"},
            },
        ),
        (
            FHIRBadRequestError(code="processing", details_text="Error"),
            status.HTTP_400_BAD_REQUEST,
            {
                "severity": "error",
                "code": "processing",
                "details": {"text": "Error"},
            },
        ),
        (
            FHIRUnauthorizedError(details_text="Error"),
            status.HTTP_401_UNAUTHORIZED,
            {
                "severity": "error",
                "code": "unknown",
                "details": {"text": "Error"},
            },
        ),
        (
            FHIRForbiddenError(details_text="Error"),
            status.HTTP_403_FORBIDDEN,
            {
                "severity": "error",
                "code": "forbidden",
                "details": {"text": "Error"},
            },
        ),
        (
            FHIRMethodNotAllowedError(details_text="Error"),
            status.HTTP_405_METHOD_NOT_ALLOWED,
            {
                "severity": "error",
                "code": "not-supported",
                "details": {"text": "Error"},
            },
        ),
        (
            FHIRNotAcceptableError(details_text="Error"),
            status.HTTP_406_NOT_ACCEPTABLE,
            {
                "severity": "error",
                "code": "not-supported",
                "details": {"text": "Error"},
            },
        ),
        (
            FHIRConflictError(details_text="Error"),
            status.HTTP_409_CONFLICT,
            {
                "severity": "error",
                "code": "conflict",
                "details": {"text": "Error"},
            },
        ),
        (
            FHIRGoneError(details_text="Error"),
            status.HTTP_410_GONE,
            {
                "severity": "error",
                "code": "deleted",
                "details": {"text": "Error"},
            },
        ),
        (
            FHIRPreconditionFailedError(details_text="Error"),
            status.HTTP_412_PRECONDITION_FAILED,
            {
                "severity": "error",
                "code": "conflict",
                "details": {"text": "Error"},
            },
        ),
        (
            FHIRUnsupportedMediaTypeError(details_text="Error"),
            status.HTTP_415_UNSUPPORTED_MEDIA_TYPE,
            {
                "severity": "error",
                "code": "not-supported",
                "details": {"text": "Error"},
            },
        ),
        (
            FHIRUnprocessableEntityError(code="invalid", details_text="Error"),
            status.HTTP_422_UNPROCESSABLE_ENTITY,
            {
                "severity": "error",
                "code": "invalid",
                "details": {"text": "Error"},
            },
        ),
    ],
    ids=[
        "http",
        "bad request",
        "unauthorized",
        "forbidden",
        "method not allowed",
        "not acceptable",
        "conflict",
        "gone",
        "precondition failed",
        "unsupported media type",
        "unprocessable entity",
    ],
)
@pytest.mark.parametrize(
    argnames="handler_func",
    argvalues=[_handler_exception_async, _handler_exception],
    ids=["async", "nonasync"],
)
def test_exception(
    exception: HTTPException,
    status_code: int,
    issue: Mapping[str, Any],
    handler_func: Callable[
        [HTTPException],
        Union[Callable[..., Coroutine[None, None, Patient]], Callable[..., Patient]],
    ],
) -> None:
    """Test exception handling for HTTP and FHIR exceptions."""
    handler = handler_func(exception)

    provider = FHIRProvider()
    provider.read(Patient)(handler)

    client = app(provider)

    response = client.get(f"/Patient/{generate_fhir_resource_id()}")

    assert_expected_response(
        response,
        status_code,
        content={
            "resourceType": "OperationOutcome",
            "issue": [issue],
        },
    )


@pytest.mark.parametrize(
    argnames="client",
    argvalues=[("read", "update", "patch", "delete", "create", "search-type")],
    ids=["all"],
    indirect=True,
)
def test_set_exception_callback(client: TestClient) -> None:
    """Test set_exception_callback."""
    test_app = cast(FHIRStarter, client.app)

    async def callback(_: Request, response_: Response, __: Exception) -> Response:
        # Change the status code from 404 to 400 so that we can test that the callback was actually
        # called.
        response_.status_code = status.HTTP_400_BAD_REQUEST
        return response_

    test_app.set_exception_callback(callback)

    id_ = generate_fhir_resource_id()
    response = client.get(f"/Patient/{id_}")

    assert_expected_response(
        response,
        status.HTTP_400_BAD_REQUEST,
        content={
            "resourceType": "OperationOutcome",
            "issue": [
                {
                    "severity": "error",
                    "code": "not-found",
                    "details": {"text": f"Unknown Patient resource '{id_}'"},
                }
            ],
        },
    )



"""Utilities for working with the FHIR specification."""

import importlib.metadata
import json
import os
import zipfile
from copy import deepcopy
from typing import Dict, Set

try:
    from functools import cache
except ImportError:
    from functools import lru_cache as cache

from pathlib import Path
from typing import Any, Mapping, Union, cast

import fhir.resources

from . import sequences

FHIR_SEQUENCE = os.getenv("FHIR_SEQUENCE", "R5")

# Set the FHIR version and the FHIR data directory location, and ensure that the specified FHIR
# sequence is supported by FHIRStarter
if FHIR_SEQUENCE == "STU3":
    import fhir.resources.STU3

    FHIR_VERSION = fhir.resources.STU3.__fhir_version__
    FHIR_DIR = Path(sequences.STU3.__file__).parent
elif FHIR_SEQUENCE == "R4":
    FHIR_VERSION = fhir.resources.__fhir_version__
    FHIR_DIR = Path(sequences.R4.__file__).parent
elif FHIR_SEQUENCE == "R4B":
    import fhir.resources.R4B

    FHIR_VERSION = fhir.resources.R4B.__fhir_version__
    FHIR_DIR = Path(sequences.R4B.__file__).parent
elif FHIR_SEQUENCE == "R5":
    FHIR_VERSION = fhir.resources.__fhir_version__
    FHIR_DIR = Path(sequences.R5.__file__).parent
else:
    raise AssertionError(f"Specified FHIR sequence must be one of: STU3, R4, R4B, R5")


# Ensure that a compatible version of fhir.resources is installed
FHIR_RESOURCES_VERSION = importlib.metadata.version("fhir.resources")
if FHIR_SEQUENCE == "R4":
    assert FHIR_RESOURCES_VERSION == "6.4.0", (
        f"fhir.resources package version must be 6.4.0 for FHIR R4 sequence; installed version is "
        f"{FHIR_RESOURCES_VERSION}"
    )
else:
    assert FHIR_RESOURCES_VERSION >= "7.0.0", (
        f"fhir.resources package version must be 7.0.0 or greater for FHIR STU3, R4B, and R5 "
        f"sequences; installed version is {FHIR_RESOURCES_VERSION}"
    )
    assert fhir.resources.__fhir_version__ == "5.0.0", (
        f"fhir.resources package references unexpected FHIR version "
        f"{fhir.resources.__fhir_version__}"
    )


def is_resource_type(resource_type: str) -> bool:
    """Return True or False depending on whether the given string is a recognized resource type."""
    return resource_type in _load_resources_list()


@cache
def _load_resources_list() -> Set[str]:
    """Load the list of resources from the JSON file."""
    with open(FHIR_DIR / "resource_types.json") as file_:
        return set(json.load(file_))


@cache
def load_examples(
    resource_type: str,
) -> Dict[str, Dict[str, Union[str, Dict[str, Any]]]]:
    """Return the examples for a specific resource type."""
    with zipfile.ZipFile(FHIR_DIR / "examples.zip") as file_:
        return json.loads(file_.read(f"{resource_type.lower()}.json"))


def create_bundle_example(resource_example: Mapping[str, Any]) -> Dict[str, Any]:
    """
    Create a bundle example for a specific resource type.

    The standard bundle example is modified based on the given resource example.
    """
    resource_type = resource_example["resourceType"]
    bundle_examples = load_examples("Bundle")
    bundle_example = deepcopy(
        cast(Dict[str, Any], next(iter(bundle_examples.values()))["value"])
    )

    bundle_example["link"][0] = {
        "relation": "self",
        "url": f"https://example.com/base/{resource_type}?_count=1",
    }
    bundle_example["link"][1] = {
        "relation": "next",
        "url": f"https://example.com/base/{resource_type}?"
        "searchId=ff15fd40-ff71-4b48-b366-09c706bed9d0&page=2",
    }
    bundle_example["entry"] = [
        {
            "fullUrl": f"https://example.com/base/{resource_type}/{resource_example['id']}",
            "resource": resource_example,
            "search": {"mode": "match", "score": 1},
        }
    ]

    return bundle_example


def make_operation_outcome_example(
    severity: str, code: str, details_text: str
) -> Dict[str, Any]:
    """Make an OperationOutcome example given a severity, code, and details text."""
    return {
        "resourceType": "OperationOutcome",
        "id": "101",
        "issue": [
            {
                "severity": severity,
                "code": code,
                "details": {"text": details_text},
            }
        ],
    }


def load_search_parameters() -> Dict[str, Any]:
    """Load the search parameters file."""
    with zipfile.ZipFile(FHIR_DIR / "search-parameters.zip") as file_:
        return json.loads(file_.read("search-parameters.json"))


def load_extra_search_parameters() -> Dict[str, Dict[str, Union[str, bool]]]:
    """Load the extra search parameters file."""
    with open(FHIR_DIR / "extra-search-parameters.json") as file_:
        return json.load(file_)



"""Classes and types for handling and representing FHIR Interactions."""

from abc import abstractmethod
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Coroutine,
    Generic,
    Literal,
    Mapping,
    Type,
    TypeVar,
    Union,
)

from fastapi import Request, Response

from .json_patch import JSONPatch
from .resources import Bundle, Id, Resource

ResourceType = TypeVar("ResourceType", bound=Resource)


@dataclass
class InteractionContext:
    request: Request
    response: Response


ReadInteractionHandler = Callable[
    [InteractionContext, Id], Union[Coroutine[None, None, ResourceType], ResourceType]
]
UpdateInteractionHandler = Callable[
    [InteractionContext, Id, ResourceType],
    Union[Coroutine[None, None, Union[Id, ResourceType]], Id, ResourceType],
]
PatchInteractionHandler = Callable[
    [InteractionContext, Id, JSONPatch],
    Union[Coroutine[None, None, Union[Id, ResourceType]], Id, ResourceType],
]
DeleteInteractionHandler = Callable[
    [InteractionContext, Id],
    Union[Coroutine[None, None, None], None],
]
CreateInteractionHandler = Callable[
    [InteractionContext, ResourceType],
    Union[Coroutine[None, None, Union[Id, ResourceType]], Id, ResourceType],
]
SearchTypeInteractionHandler = Callable[
    ..., Union[Coroutine[None, None, Bundle], Bundle]
]

InteractionHandler = Union[
    ReadInteractionHandler[ResourceType],
    PatchInteractionHandler[ResourceType],
    DeleteInteractionHandler,
    UpdateInteractionHandler[ResourceType],
    CreateInteractionHandler[ResourceType],
    SearchTypeInteractionHandler,
]


class TypeInteraction(Generic[ResourceType]):
    """
    Collection of values that represent a FHIR type interactions. This class can also represent
    instance level interactions.

    resource_type:    The type of FHIR resource on which this interaction operates, as defined by
                      the fhir.resources package.
    handler:          User-defined function that performs the FHIR interaction.
    route_options:    Dictionary of key-value pairs that are passed on to FastAPI on route creation.
    """

    def __init__(
        self,
        resource_type: Type[ResourceType],
        handler: InteractionHandler[ResourceType],
        route_options: Mapping[str, Any],
    ):
        self.resource_type = resource_type
        self.handler = handler
        self.route_options = route_options

    @staticmethod
    @abstractmethod
    def label() -> str:
        raise NotImplementedError


class ReadInteraction(TypeInteraction[ResourceType]):
    @staticmethod
    def label() -> Literal["read"]:
        return "read"


class UpdateInteraction(TypeInteraction[ResourceType]):
    @staticmethod
    def label() -> Literal["update"]:
        return "update"


class PatchInteraction(TypeInteraction[ResourceType]):
    @staticmethod
    def label() -> Literal["patch"]:
        return "patch"


class DeleteInteraction(TypeInteraction[ResourceType]):
    @staticmethod
    def label() -> Literal["delete"]:
        return "delete"


class CreateInteraction(TypeInteraction[ResourceType]):
    @staticmethod
    def label() -> Literal["create"]:
        return "create"


class SearchTypeInteraction(TypeInteraction[ResourceType]):
    @staticmethod
    def label() -> Literal["search-type"]:
        return "search-type"



import re
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    MutableSequence,
    Optional,
)

from pydantic import BaseModel, Extra, Field, ValidationError, root_validator, validator
from pydantic.error_wrappers import ErrorWrapper

_PATH_PATTERN = re.compile("^\/(?:[^/]+\/)*[^/]+$")


class JSONPatchOperation(BaseModel):
    """
    JSON Patch operation Pydantic model (https://datatracker.ietf.org/doc/html/rfc6902).
    """

    op: Literal["add", "remove", "replace", "move", "copy", "test"]
    from_: Optional[str] = Field(alias="from")
    path: str
    value: Optional[Any]

    class Config:
        extra = Extra.forbid
        schema_extra = {
            "example": {
                "op": "add",
                "path": "/text",
                "value": {
                    "status": "empty",
                    "div": '<div xmlns="http://www.w3.org/1999/xhtml">No human-readable text provided in this case</div>',
                },
            }
        }

    @validator("from_", "path")
    def validate_json_pointers(cls, json_pointer: str) -> str:
        """Ensure that the from and path fields contain valid JSON Pointers."""
        if not re.fullmatch(_PATH_PATTERN, json_pointer):
            raise ValueError(f"invalid JSON Pointer")
        return json_pointer

    @root_validator(pre=False, skip_on_failure=True)
    def validate_structure(cls, values: Mapping[str, Any]) -> Mapping[str, Any]:
        """
        Ensure that this operation has the correct fields based on the operation type (e.g. add,
        replace, etc.)
        """
        errors: List[ErrorWrapper] = []

        op = values["op"]

        cls._check_optional_field(
            op,
            field="from",
            value=values.get("from_"),
            allowed_ops=("move", "copy"),
            errors=errors,
        )
        cls._check_optional_field(
            op,
            field="value",
            value=values.get("value"),
            allowed_ops=("add", "replace", "test"),
            errors=errors,
        )

        if errors:
            raise ValidationError(errors, cls)

        return values

    @staticmethod
    def _check_optional_field(
        op: str,
        field: str,
        value: Any,
        allowed_ops: Iterable[str],
        errors: MutableSequence[ErrorWrapper],
    ) -> None:
        if value and op not in allowed_ops:
            errors.append(
                ErrorWrapper(ValueError("extra fields not permitted"), loc=(field,))
            )
        if not value and op in allowed_ops:
            errors.append(ErrorWrapper(ValueError("field required"), loc=(field,)))


JSONPatch = List[JSONPatchOperation]


def convert_json_patch(json_patch: JSONPatch) -> List[Dict[str, Any]]:
    """
    Convert the JSON Patch object to a list of dicts.

    This method will also rename the "from_" key in the operation dictionaries to "from".
    """
    return [
        {
            key.replace("from_", "from"): value
            for key, value in op.dict().items()
            if value
        }
        for op in json_patch
    ]
